import numpy as np
from sklearn.base import BaseEstimator
from sklearn.feature_selection._base import SelectorMixin

class FeatureEncodingFrequencySelector(BaseEstimator, SelectorMixin):
    """Feature selector based on Encoding Frequency. Encoding frequency is the frequency of each unique element(0/1/2/3) present in a feature set. 
     Features are selected on the basis of a threshold assigned for encoding frequency. If frequency of any unique element is less than or equal to threshold, the feature is removed.  """

    @property
    def __name__(self):
        """Instance name is the same as the class name. """
        return self.__class__.__name__
    
    def __init__(self, threshold):
        """Create a FeatureEncodingFrequencySelector object.
        
        Parameters
        ----------
        threshold : float, required
            Threshold value for allele frequency. If frequency of A or frequency of a is less than the threshold value then the feature is dropped.
            
        Returns
        -------
        None
        
        """
        self.threshold = threshold

    """def fit(self, X, y=None):
        Fit FeatureAlleleFrequencySelector for feature selection
        
        Parameters
        ----------
        X : numpy ndarray, {n_samples, n_features}
            The training input samples.
        y : numpy array {n_samples,}
            The training target values.
            
        Returns
        -------
        self : object
            Returns a copy of the estimator
        
        self.selected_feature_indexes = []
        self.no_of_features = X.shape[1]

        # Finding the no of alleles in each feature column
        for i in range(0, X.shape[1]):
            no_of_AA_featurewise = np.count_nonzero(X[:,i]==0)
            no_of_Aa_featurewise = np.count_nonzero(X[:,i]==1)
            no_of_aa_featurewise = np.count_nonzero(X[:,i]==2)
    
    
            frequency_A_featurewise = (2*no_of_AA_featurewise + no_of_Aa_featurewise) / (2*no_of_AA_featurewise + 
            2*no_of_Aa_featurewise + 2*no_of_aa_featurewise)

            frequency_a_featurewise = 1 - frequency_A_featurewise

            if(not(frequency_A_featurewise <= self.threshold) and not(frequency_a_featurewise <= self.threshold)):
                self.selected_feature_indexes.append(i)
        return self"""

    """def transform(self, X):
        Make subset after fit
        
        Parameters
        ----------
        X : numpy ndarray, {n_samples, n_features}
            New data, where n_samples is the number of samples and n_features is the number of features.
            
        Returns
        -------
        X_transformed : numpy ndarray, {n_samples, n_features}
            The transformed feature set.
        
        
        X_transformed = X[:, self.selected_feature_indexes]

        return X_transformed"""

    def fit(self, X, y=None) :
        """Fit FeatureEncodingFrequencySelector for feature selection. This function gets the appropriate features. """
       
        self.selected_feature_indexes = []
        self.no_of_original_features = X.shape[1]

        # Finding the frequency of all the unique elements present featurewise in the input variable X
        for i in range(0, X.shape[1]):
            unique, counts = np.unique(X[:,i], return_counts=True)
            element_count_dict_featurewise = dict(zip(unique, counts))
            element_frequency_dict_featurewise = {}
            feature_column_selected = True

            for x in unique:
                x_frequency_featurewise = element_count_dict_featurewise[x] / sum(counts)
                element_frequency_dict_featurewise[x] = x_frequency_featurewise
            
            for frequency in element_frequency_dict_featurewise.values():
                if frequency <= self.threshold :
                    feature_column_selected = False
                    break
            
            if feature_column_selected == True :
                self.selected_feature_indexes.append(i)
        
        if not len(self.selected_feature_indexes):
            """msg = "No feature in X meets the encoding frequency threshold {0:.5f}"
            raise ValueError(msg.format(self.threshold))"""
            for i in range(0, X.shape[1]):
                self.selected_feature_indexes.append(i)
        
        return self

    def transform(self, X):
        """ Make subset after fit. This function returns a transformed version of X.  """
        X_transformed = X[:, self.selected_feature_indexes]

        return X_transformed


    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected
        It is the abstractmethod
        
        Returns
        -------
        support : boolean array of shape [# input features]
            An element is True iff its corresponding feature is selected for retention.
            """
        n_features = self.no_of_original_features
        mask = np.zeros(n_features, dtype=bool)
        mask[np.asarray(self.selected_feature_indexes)] = True

        return mask



from functools import partial
import numpy as np

from tpot2.builtin_modules import ZeroCount, OneHotEncoder, ColumnOneHotEncoder
from sklearn.preprocessing import Binarizer
from sklearn.decomposition import FastICA
from sklearn.cluster import FeatureAgglomeration
from sklearn.preprocessing import MaxAbsScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import Normalizer
from sklearn.kernel_approximation import Nystroem
from sklearn.decomposition import PCA
from sklearn.preprocessing import PolynomialFeatures
from sklearn.kernel_approximation import RBFSampler
from sklearn.preprocessing import RobustScaler
from sklearn.preprocessing import StandardScaler


def params_sklearn_preprocessing_Binarizer(trial, name=None):
    return {
        'threshold': trial.suggest_float(f'threshold_{name}', 0.0, 1.0),
    }

def params_sklearn_decomposition_FastICA(trial, random_state=None, name=None, n_features=100):
    return {
        'n_components': trial.suggest_int(f'n_components_{name}', 1, n_features), # number of components wrt number of features
        'algorithm': trial.suggest_categorical(f'algorithm_{name}', ['parallel', 'deflation']),
        'whiten':'unit-variance',
        'random_state': random_state
    }

def params_sklearn_cluster_FeatureAgglomeration(trial, name=None, n_features=100):

    linkage = trial.suggest_categorical(f'linkage_{name}', ['ward', 'complete', 'average'])
    if linkage == 'ward':
        metric = 'euclidean'
    else:
        metric = trial.suggest_categorical(f'metric_{name}', ['euclidean', 'l1', 'l2', 'manhattan', 'cosine'])
    return {
        'linkage': linkage,
        'metric': metric,
        'n_clusters': trial.suggest_int(f'n_clusters_{name}', 2, n_features-1), #TODO perhaps a percentage of n_features
    }

def params_sklearn_preprocessing_Normalizer(trial, name=None):
    return {
        'norm': trial.suggest_categorical(f'norm_{name}', ['l1', 'l2', 'max']),
    }

def params_sklearn_kernel_approximation_Nystroem(trial, random_state=None, name=None, n_features=100):
    return {
        'gamma': trial.suggest_float(f'gamma_{name}', 0.0, 1.0),
        'kernel': trial.suggest_categorical(f'kernel_{name}', ['rbf', 'cosine', 'chi2', 'laplacian', 'polynomial', 'poly', 'linear', 'additive_chi2', 'sigmoid']),
        'n_components': trial.suggest_int(f'n_components_{name}', 1, n_features),
        'random_state': random_state
    }

def params_sklearn_decomposition_PCA(trial, random_state=None, name=None, n_features=100):
    # keep the number of components required to explain 'variance_explained' of the variance
    variance_explained = 1.0 - trial.suggest_float(f'n_components_{name}', 0.001, 0.5, log=True) #values closer to 1 are more likely

    return {
        'n_components': variance_explained,
        'random_state': random_state
    }

def params_sklearn_kernel_approximation_RBFSampler(trial, random_state=None, name=None, n_features=100):
    return {
        'n_components': trial.suggest_int(f'n_components_{name}', 1, n_features),
        'gamma': trial.suggest_float(f'gamma_{name}', 0.0, 1.0),
        'random_state': random_state
    }

def params_tpot_builtins_ZeroCount(trial, name=None):

    return {}

def params_tpot_builtins_OneHotEncoder(trial, name=None):

    return {}

def make_transformer_config_dictionary(random_state=None, n_features=10):
    #n_features = min(n_features,100) #TODO optimize this
    return {
                Binarizer: params_sklearn_preprocessing_Binarizer,
                FastICA: partial(params_sklearn_decomposition_FastICA, random_state=random_state, n_features=n_features),
                FeatureAgglomeration: partial(params_sklearn_cluster_FeatureAgglomeration,n_features=n_features),
                MaxAbsScaler: {},
                MinMaxScaler: {},
                Normalizer: params_sklearn_preprocessing_Normalizer,
                Nystroem: partial(params_sklearn_kernel_approximation_Nystroem, random_state=random_state, n_features=n_features),
                PCA: partial(params_sklearn_decomposition_PCA, random_state=random_state, n_features=n_features),
                PolynomialFeatures: {
                                        'degree': 2,
                                        'include_bias': False,
                                        'interaction_only': False,
                                    },
                RBFSampler: partial(params_sklearn_kernel_approximation_RBFSampler, random_state=random_state, n_features=n_features),
                RobustScaler: {},
                StandardScaler: {},
                ZeroCount: params_tpot_builtins_ZeroCount,
                ColumnOneHotEncoder: params_tpot_builtins_OneHotEncoder,
            }



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""This file is part of the TPOT library.

TPOT was primarily developed at the University of Pennsylvania by:
    - Randal S. Olson (rso@randalolson.com)
    - Weixuan Fu (weixuanf@upenn.edu)
    - Daniel Angell (dpa34@drexel.edu)
    - and many more generous open source contributors

TPOT is free software: you can redistribute it and/or modify
it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3 of
the License, or (at your option) any later version.

TPOT is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with TPOT. If not, see <http://www.gnu.org/licenses/>.
"""
#TODO handle sparse input?

import numpy as np
import pandas as pd
import os, os.path
from sklearn.base import BaseEstimator
from sklearn.feature_selection._base import SelectorMixin



#TODO clean this up and make sure it works
class FeatureSetSelector(BaseEstimator, SelectorMixin):
    """Select predefined feature subsets."""

    def __init__(self, sel_subset=None, name=None):
        """Create a FeatureSetSelector object.

        Parameters
        ----------
        sel_subset: list or int
            If X is a dataframe, items in sel_subset list must correspond to column names
            If X is a numpy array, items in sel_subset list must correspond to column indexes
            int: index of a single column
        Returns
        -------
        None

        """
        self.name = name
        self.sel_subset = sel_subset

        
    def fit(self, X, y=None):
        """Fit FeatureSetSelector for feature selection

        Parameters
        ----------
        X: array-like of shape (n_samples, n_features)
            The training input samples.
        y: array-like, shape (n_samples,)
            The target values (integers that correspond to classes in classification, real numbers in regression).

        Returns
        -------
        self: object
            Returns a copy of the estimator
        """
        if isinstance(self.sel_subset, int) or isinstance(self.sel_subset, str):
            self.sel_subset = [self.sel_subset]

        #generate  self.feat_list_idx
        if isinstance(X, pd.DataFrame):
            self.feature_names_in_ = X.columns.tolist()
            self.feat_list_idx = sorted([self.feature_names_in_.index(feat) for feat in self.sel_subset])

            
        elif isinstance(X, np.ndarray):
            self.feature_names_in_ = None#list(range(X.shape[1]))

            self.feat_list_idx = sorted(self.sel_subset)
        
        n_features = X.shape[1]
        self.mask = np.zeros(n_features, dtype=bool)
        self.mask[np.asarray(self.feat_list_idx)] = True

        return self

    #TODO keep returned as dataframe if input is dataframe? may not be consistent with sklearn

    # def transform(self, X):
    

    def _get_support_mask(self):
        """
        Get the boolean mask indicating which features are selected
        Returns
        -------
        support : boolean array of shape [# input features]
            An element is True iff its corresponding feature is selected for
            retention.
        """
        return self.mask




import networkx as nx
import numpy as np


def remove_and_stitch(graph, node):
    successors = graph.successors(node)
    predecessors = graph.predecessors(node)

    graph.remove_node(node)

    for s in successors:
        for p in predecessors:
            graph.add_edge(p,s)


def remove_nodes_disconnected_from_node(graph, node):
    descendants = nx.descendants(graph, node)
    for n in list(graph.nodes):
        if n not in descendants and n is not node:
            graph.remove_node(n)
    #graph.remove_nodes_from([n for n in graph.nodes if n not in nx.descendants(graph, node) and n is not node])


def get_roots(graph):
    return [v for v, d in graph.in_degree() if d == 0]

def get_leaves(graph):
    return [v for v, d in graph.out_degree() if d == 0]

def get_max_path_through_node(graph, root, node):
    if len(list(graph.successors(node)))==0:
        return get_max_path_size(graph, root, node)
    else:
        leaves = [n for n in nx.descendants(graph,node) if len(list(graph.successors(n)))==0]

        return max([get_max_path_size(graph, root, l) for l in leaves])


def get_max_path_size(graph, fromnode1,tonode2, return_path=False):
    if fromnode1 is tonode2:
        if return_path:
            return [fromnode1]
        return 1
    else:
        max_length_path = max(nx.all_simple_paths(graph, fromnode1, tonode2), key=lambda x: len(x))
        if return_path:
            return max_length_path
        return len(max_length_path) #gets the max path and finds the length of that path


def invert_dictionary(d):
    inv_map = {}
    for k, v in d.items():
        inv_map.setdefault(v, set()).add(k)

    return inv_map

def select_nodes_same_depth(g1, node1, g2, node2, rng_=None):
    rng = np.random.default_rng(rng_)

    g1_nodes = nx.shortest_path_length(g1, source=node1)
    g2_nodes = nx.shortest_path_length(g2, source=node2)

    max_depth = max(list(g1_nodes.values()) + list(g2_nodes.values()))

    g1_nodes = invert_dictionary(g1_nodes)
    g2_nodes = invert_dictionary(g2_nodes)

    # depth_number_of_nodes = []
    # for i in range(max_depth+1):
    #     n = 0
    #     if i in g1_nodes and i in g2_nodes:
    #         depth_number_of_nodes.append(len(g1_nodes[i])+len(g1_nodes[i]))
    #     else:
    #         break

    possible_pairs = []
    for i in range(max_depth+1):
        if i in g1_nodes and i in g2_nodes:
            for n1 in g1_nodes[i]:
                for n2 in g2_nodes[i]:
                    possible_pairs.append( (n1,n2) )

    rng.shuffle(possible_pairs)

    for p in possible_pairs:
        yield p[0], p[1]

def select_nodes_randomly(g1, g2, rng_=None):
    rng = np.random.default_rng(rng_)

    sorted_self_nodes_list = list(g1.nodes)
    rng.shuffle(sorted_self_nodes_list)

    sorted_other_nodes_list = list(g2.nodes)
    rng.shuffle(sorted_other_nodes_list)
    for node1 in sorted_self_nodes_list:
        for node2 in sorted_other_nodes_list:
            yield node1, node2


# Test all nodes have all dictionaries
import pytest
import tpot2.graphsklearn as GraphSklearn


@pytest.mark.skip(reason="Not yet implemented")
def test_BaseNode_static_methods():
    n1 = GraphSklearn.BaseNode()
    n2 = GraphSklearn.BaseNode()
    n3 = GraphSklearn.BaseNode()
    n4 = GraphSklearn.BaseNode()
    n5 = GraphSklearn.BaseNode()
    n6 = GraphSklearn.BaseNode()
    n7 = GraphSklearn.BaseNode()

    GraphSklearn.BaseNode._connect_nodes(child=n2,parent=n1)
    GraphSklearn.BaseNode._connect_nodes(child=n3,parent=n1)
    GraphSklearn.BaseNode._connect_nodes(child=n4,parent=n1)

    GraphSklearn.BaseNode._connect_nodes(child=n5,parent=n2)

    GraphSklearn.BaseNode._connect_nodes(child=n5,parent=n3)
    GraphSklearn.BaseNode._connect_nodes(child=n6,parent=n3)

    GraphSklearn.BaseNode._connect_nodes(child=n6,parent=n4)
    GraphSklearn.BaseNode._connect_nodes(child=n3,parent=n4)

    GraphSklearn.BaseNode._connect_nodes(child=n4,parent=n7) #This is a second root, allowed by basenode

    assert n2 in n1.children
    assert n3 in n1.children
    assert n4 in n1.children
    assert n5 in n2.children
    assert n5 in n3.children
    assert n6 in n3.children
    assert n6 in n4.children
    assert n3 in n4.children
    assert n4 in n7.children

    assert n1 in n2.parents
    assert n1 in n3.parents
    assert n1 in n4.parents
    assert n2 in n5.parents
    assert n3 in n5.parents
    assert n3 in n6.parents
    assert n4 in n6.parents
    assert n4 in n3.parents
    assert n7 in n4.parents

    assert n2 not in n1.parents
    assert n3 not in n1.parents
    assert n4 not in n1.parents
    assert n6 not in n2.parents
    assert n5 not in n3.parents
    assert n6 not in n3.parents
    assert n6 not in n4.parents

    #Make sure every node knows about every other node in the graph.
    for this_node in [n1,n2,n3,n4,n5,n6, n7]:
        for other_node in [n1,n2,n3,n4,n5,n6, n7]:
            assert this_node in other_node.node_set


    GraphSklearn.BaseNode._disconnect_nodes(child=n3, parent=n4)
    assert n4 not in n3.parents
    assert n3 not in n4.children

    n8 = GraphSklearn.BaseNode()
    GraphSklearn.BaseNode._insert_inner_node(child=n3,new_node=n8,parent=n4)
    assert n8 in n3.parents
    assert n4 in n8.parents

    assert n3 in n8.children
    assert n8 in n4.children

    n9 = GraphSklearn.BaseNode()

    GraphSklearn.BaseNode._remove_node(node=n8, replacement=n9)
    assert n8 not in n3.parents
    assert n4 not in n8.parents

    assert n3 not in n8.children
    assert n8 not in n4.children
    assert n8 not in n1.node_set

    assert n9 in n3.parents
    assert n4 in n9.parents

    assert n3 in n9.children
    assert n9 in n4.children

    for this_node in [n1,n2,n3,n4,n5,n6, n7, n9]:
        for other_node in [n1,n2,n3,n4,n5,n6, n7, n9]:
            assert this_node in other_node.node_set


@pytest.mark.skip(reason="Not yet implemented")
def test_BaseNode_static_crossover_methods():
    n1 = GraphSklearn.BaseNode()
    n2 = GraphSklearn.BaseNode()
    n3 = GraphSklearn.BaseNode()
    n4 = GraphSklearn.BaseNode()
    n5 = GraphSklearn.BaseNode()
    n6 = GraphSklearn.BaseNode()
    n7 = GraphSklearn.BaseNode()
    n8 = GraphSklearn.BaseNode()

    GraphSklearn.BaseNode._connect_nodes(child=n2,parent=n1)
    GraphSklearn.BaseNode._connect_nodes(child=n3,parent=n1)
    GraphSklearn.BaseNode._connect_nodes(child=n4,parent=n1)

    GraphSklearn.BaseNode._connect_nodes(child=n5,parent=n2)

    GraphSklearn.BaseNode._connect_nodes(child=n5,parent=n3)
    GraphSklearn.BaseNode._connect_nodes(child=n6,parent=n3)

    GraphSklearn.BaseNode._connect_nodes(child=n6,parent=n4)
    GraphSklearn.BaseNode._connect_nodes(child=n3,parent=n4)

    GraphSklearn.BaseNode._connect_nodes(child=n4,parent=n7) #This is a second root, allowed by basenode

    GraphSklearn.BaseNode._connect_nodes(child=n8,parent=n5)
    GraphSklearn.BaseNode._connect_nodes(child=n8,parent=n1)

    depth_dict = n1._get_depth_dictionary()

    assert depth_dict[n1] == 0
    assert depth_dict[n2] == 1
    assert depth_dict[n3] == 2
    assert depth_dict[n4] == 1
    assert depth_dict[n5] == 3
    assert depth_dict[n6] == 3
    assert depth_dict[n8] == 4
    assert n7 not in depth_dict


from sklearn.base import BaseEstimator
from sklearn.utils.metaestimators import available_if
import numpy as np
import sklearn.metrics
import tpot2.config
from sklearn.utils.validation import  check_is_fitted
from tpot2.selectors import survival_select_NSGA2, tournament_selection_dominated
from sklearn.preprocessing import LabelEncoder
from sklearn.utils.multiclass import unique_labels
import pandas as pd
from sklearn.model_selection import train_test_split
import tpot2

from dask.distributed import Client
from dask.distributed import LocalCluster
import math

from dask import config as cfg
from .estimator_utils import *
import warnings

def set_dask_settings():
    cfg.set({'distributed.scheduler.worker-ttl': None})
    cfg.set({'distributed.scheduler.allowed-failures':1})



#TODO inherit from _BaseComposition?
class TPOTEstimatorSteadyState(BaseEstimator):
    def __init__(self,  scorers= [],
                        scorers_weights = [],
                        classification = False,
                        cv = 5,
                        other_objective_functions=[], #tpot2.objectives.estimator_objective_functions.number_of_nodes_objective],
                        other_objective_functions_weights = [],
                        objective_function_names = None,
                        bigger_is_better = True,
                        max_size = np.inf,
                        linear_pipeline = False,
                        root_config_dict= 'Auto',
                        inner_config_dict=["selectors", "transformers"],
                        leaf_config_dict= None,
                        cross_val_predict_cv = 0,
                        categorical_features = None,
                        subsets = None,
                        memory = None,
                        preprocessing = False,
                        validation_strategy = "none",
                        validation_fraction = .2,
                        disable_label_encoder = False,

                        initial_population_size = 50,
                        population_size = 50,
                        max_evaluated_individuals = None,



                        early_stop = None,
                        early_stop_seconds = None,
                        scorers_early_stop_tol = 0.001,
                        other_objectives_early_stop_tol = None,
                        max_time_seconds=None,
                        max_eval_time_seconds=60*10,
                        n_jobs=1,
                        memory_limit = "4GB",
                        client = None,

                        crossover_probability=.2,
                        mutate_probability=.7,
                        mutate_then_crossover_probability=.05,
                        crossover_then_mutate_probability=.05,
                        survival_selector = survival_select_NSGA2,
                        parent_selector = tournament_selection_dominated,
                        budget_range = None,
                        budget_scaling = .5,
                        individuals_until_end_budget = 1,
                        stepwise_steps = 5,

                        warm_start = False,
                        subset_column = None,

                        verbose = 0,
                        periodic_checkpoint_folder = None,
                        callback = None,
                        processes = True,

                        scatter = True,

                        # random seed for random number generator (rng)
                        random_state = None,

                        optuna_optimize_pareto_front = False,
                        optuna_optimize_pareto_front_trials = 100,
                        optuna_optimize_pareto_front_timeout = 60*10,
                        optuna_storage = "sqlite:///optuna.db",
                        ):

        '''
        An sklearn baseestimator that uses genetic programming to optimize a pipeline.

        Parameters
        ----------

        scorers : (list, scorer)
            A scorer or list of scorers to be used in the cross-validation process.
            see https://scikit-learn.org/stable/modules/model_evaluation.html

        scorers_weights : list
            A list of weights to be applied to the scorers during the optimization process.

        classification : bool
            If True, the problem is treated as a classification problem. If False, the problem is treated as a regression problem.
            Used to determine the CV strategy.

        cv : int, cross-validator
            - (int): Number of folds to use in the cross-validation process. By uses the sklearn.model_selection.KFold cross-validator for regression and StratifiedKFold for classification. In both cases, shuffled is set to True.
            - (sklearn.model_selection.BaseCrossValidator): A cross-validator to use in the cross-validation process.

        other_objective_functions : list, default=[]
            A list of other objective functions to apply to the pipeline. The function takes a single parameter for the graphpipeline estimator and returns either a single score or a list of scores.

        other_objective_functions_weights : list, default=[]
            A list of weights to be applied to the other objective functions.

        objective_function_names : list, default=None
            A list of names to be applied to the objective functions. If None, will use the names of the objective functions.

        bigger_is_better : bool, default=True
            If True, the objective function is maximized. If False, the objective function is minimized. Use negative weights to reverse the direction.


        max_size : int, default=np.inf
            The maximum number of nodes of the pipelines to be generated.

        linear_pipeline : bool, default=False
            If True, the pipelines generated will be linear. If False, the pipelines generated will be directed acyclic graphs.

        root_config_dict : dict, default='auto'
            The configuration dictionary to use for the root node of the model.
            If 'auto', will use "classifiers" if classification=True, else "regressors".
            - 'selectors' : A selection of sklearn Selector methods.
            - 'classifiers' : A selection of sklearn Classifier methods.
            - 'regressors' : A selection of sklearn Regressor methods.
            - 'transformers' : A selection of sklearn Transformer methods.
            - 'arithmetic_transformer' : A selection of sklearn Arithmetic Transformer methods that replicate symbolic classification/regression operators.
            - 'passthrough' : A node that just passes though the input. Useful for passing through raw inputs into inner nodes.
            - 'feature_set_selector' : A selector that pulls out specific subsets of columns from the data. Only well defined as a leaf.
                                        Subsets are set with the subsets parameter.
            - 'skrebate' : Includes ReliefF, SURF, SURFstar, MultiSURF.
            - 'MDR' : Includes MDR.
            - 'ContinuousMDR' : Includes ContinuousMDR.
            - 'genetic encoders' : Includes Genetic Encoder methods as used in AutoQTL.
            - 'FeatureEncodingFrequencySelector': Includes FeatureEncodingFrequencySelector method as used in AutoQTL.
            - list : a list of strings out of the above options to include the corresponding methods in the configuration dictionary.

        inner_config_dict : dict, default=["selectors", "transformers"]
            The configuration dictionary to use for the inner nodes of the model generation.
            Default ["selectors", "transformers"]
            - 'selectors' : A selection of sklearn Selector methods.
            - 'classifiers' : A selection of sklearn Classifier methods.
            - 'regressors' : A selection of sklearn Regressor methods.
            - 'transformers' : A selection of sklearn Transformer methods.
            - 'arithmetic_transformer' : A selection of sklearn Arithmetic Transformer methods that replicate symbolic classification/regression operators.
            - 'passthrough' : A node that just passes though the input. Useful for passing through raw inputs into inner nodes.
            - 'feature_set_selector' : A selector that pulls out specific subsets of columns from the data. Only well defined as a leaf.
                                        Subsets are set with the subsets parameter.
            - 'skrebate' : Includes ReliefF, SURF, SURFstar, MultiSURF.
            - 'MDR' : Includes MDR.
            - 'ContinuousMDR' : Includes ContinuousMDR.
            - 'genetic encoders' : Includes Genetic Encoder methods as used in AutoQTL.
            - 'FeatureEncodingFrequencySelector': Includes FeatureEncodingFrequencySelector method as used in AutoQTL.
            - list : a list of strings out of the above options to include the corresponding methods in the configuration dictionary.
            - None : If None and max_depth>1, the root_config_dict will be used for the inner nodes as well.

        leaf_config_dict : dict, default=None
            The configuration dictionary to use for the leaf node of the model. If set, leaf nodes must be from this dictionary.
            Otherwise leaf nodes will be generated from the root_config_dict.
            Default None
            - 'selectors' : A selection of sklearn Selector methods.
            - 'classifiers' : A selection of sklearn Classifier methods.
            - 'regressors' : A selection of sklearn Regressor methods.
            - 'transformers' : A selection of sklearn Transformer methods.
            - 'arithmetic_transformer' : A selection of sklearn Arithmetic Transformer methods that replicate symbolic classification/regression operators.
            - 'passthrough' : A node that just passes though the input. Useful for passing through raw inputs into inner nodes.
            - 'feature_set_selector' : A selector that pulls out specific subsets of columns from the data. Only well defined as a leaf.
                                        Subsets are set with the subsets parameter.
            - 'skrebate' : Includes ReliefF, SURF, SURFstar, MultiSURF.
            - 'MDR' : Includes MDR.
            - 'ContinuousMDR' : Includes ContinuousMDR.
            - 'genetic encoders' : Includes Genetic Encoder methods as used in AutoQTL.
            - 'FeatureEncodingFrequencySelector': Includes FeatureEncodingFrequencySelector method as used in AutoQTL.
            - list : a list of strings out of the above options to include the corresponding methods in the configuration dictionary.
            - None : If None, a leaf will not be required (i.e. the pipeline can be a single root node). Leaf nodes will be generated from the inner_config_dict.

        cross_val_predict_cv : int, default=0
            Number of folds to use for the cross_val_predict function for inner classifiers and regressors. Estimators will still be fit on the full dataset, but the following node will get the outputs from cross_val_predict.

            - 0-1 : When set to 0 or 1, the cross_val_predict function will not be used. The next layer will get the outputs from fitting and transforming the full dataset.
            - >=2 : When fitting pipelines with inner classifiers or regressors, they will still be fit on the full dataset.
                    However, the output to the next node will come from cross_val_predict with the specified number of folds.

        categorical_features: list or None
            Categorical columns to inpute and/or one hot encode during the preprocessing step. Used only if preprocessing is not False.
            - None : If None, TPOT2 will automatically use object columns in pandas dataframes as objects for one hot encoding in preprocessing.
            - List of categorical features. If X is a dataframe, this should be a list of column names. If X is a numpy array, this should be a list of column indices

        subsets : str or list, default=None
            Sets the subsets that the FeatureSetSeletor will select from if set as an option in one of the configuration dictionaries.
            - str : If a string, it is assumed to be a path to a csv file with the subsets.
                The first column is assumed to be the name of the subset and the remaining columns are the features in the subset.
            - list or np.ndarray : If a list or np.ndarray, it is assumed to be a list of subsets.
            - None : If None, each column will be treated as a subset. One column will be selected per subset.
            If subsets is None, each column will be treated as a subset. One column will be selected per subset.


        memory: Memory object or string, default=None
            If supplied, pipeline will cache each transformer after calling fit. This feature
            is used to avoid computing the fit transformers within a pipeline if the parameters
            and input data are identical with another fitted pipeline during optimization process.
            - String 'auto':
                TPOT uses memory caching with a temporary directory and cleans it up upon shutdown.
            - String path of a caching directory
                TPOT uses memory caching with the provided directory and TPOT does NOT clean
                the caching directory up upon shutdown. If the directory does not exist, TPOT will
                create it.
            - Memory object:
                TPOT uses the instance of joblib.Memory for memory caching,
                and TPOT does NOT clean the caching directory up upon shutdown.
            - None:
                TPOT does not use memory caching.

        preprocessing : bool or BaseEstimator/Pipeline,
            EXPERIMENTAL
            A pipeline that will be used to preprocess the data before CV.
            - bool : If True, will use a default preprocessing pipeline.
            - Pipeline : If an instance of a pipeline is given, will use that pipeline as the preprocessing pipeline.

        validation_strategy : str, default='none'
            EXPERIMENTAL The validation strategy to use for selecting the final pipeline from the population. TPOT2 may overfit the cross validation score. A second validation set can be used to select the final pipeline.
            - 'auto' : Automatically determine the validation strategy based on the dataset shape.
            - 'reshuffled' : Use the same data for cross validation and final validation, but with different splits for the folds. This is the default for small datasets.
            - 'split' : Use a separate validation set for final validation. Data will be split according to validation_fraction. This is the default for medium datasets.
            - 'none' : Do not use a separate validation set for final validation. Select based on the original cross-validation score. This is the default for large datasets.

        validation_fraction : float, default=0.2
          EXPERIMENTAL The fraction of the dataset to use for the validation set when validation_strategy is 'split'. Must be between 0 and 1.

        disable_label_encoder : bool, default=False
            If True, TPOT will check if the target needs to be relabeled to be sequential ints from 0 to N. This is necessary for XGBoost compatibility. If the labels need to be encoded, TPOT2 will use sklearn.preprocessing.LabelEncoder to encode the labels. The encoder can be accessed via the self.label_encoder_ attribute.
            If False, no additional label encoders will be used.

        population_size : int, default=50
            Size of the population

        initial_population_size : int, default=None
            Size of the initial population. If None, population_size will be used.

        population_scaling : int, default=0.5
            Scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.

        generations_until_end_population : int, default=1
            Number of generations until the population size reaches population_size

        generations : int, default=50
            Number of generations to run

        early_stop : int, default=None
            Number of evaluated individuals without improvement before early stopping. Counted across all objectives independently. Triggered when all objectives have not improved by the given number of individuals.

        early_stop_seconds : float, default=None
            Number of seconds without improvement before early stopping. All objectives must not have improved for the given number of seconds for this to be triggered.

        scorers_early_stop_tol :
            -list of floats
                list of tolerances for each scorer. If the difference between the best score and the current score is less than the tolerance, the individual is considered to have converged
                If an index of the list is None, that item will not be used for early stopping
            -int
                If an int is given, it will be used as the tolerance for all objectives

        other_objectives_early_stop_tol :
            -list of floats
                list of tolerances for each of the other objective function. If the difference between the best score and the current score is less than the tolerance, the individual is considered to have converged
                If an index of the list is None, that item will not be used for early stopping
            -int
                If an int is given, it will be used as the tolerance for all objectives

        max_time_seconds : float, default=float("inf")
            Maximum time to run the optimization. If none or inf, will run until the end of the generations.

        max_eval_time_seconds : float, default=60*5
            Maximum time to evaluate a single individual. If none or inf, there will be no time limit per evaluation.

        n_jobs : int, default=1
            Number of processes to run in parallel.

        memory_limit : str, default="4GB"
            Memory limit for each job. See Dask [LocalCluster documentation](https://distributed.dask.org/en/stable/api.html#distributed.Client) for more information.

        client : dask.distributed.Client, default=None
            A dask client to use for parallelization. If not None, this will override the n_jobs and memory_limit parameters. If None, will create a new client with num_workers=n_jobs and memory_limit=memory_limit.

        crossover_probability : float, default=.2
            Probability of generating a new individual by crossover between two individuals.

        mutate_probability : float, default=.7
            Probability of generating a new individual by crossover between one individuals.

        mutate_then_crossover_probability : float, default=.05
            Probability of generating a new individual by mutating two individuals followed by crossover.

        crossover_then_mutate_probability : float, default=.05
            Probability of generating a new individual by crossover between two individuals followed by a mutation of the resulting individual.

        survival_selector : function, default=survival_select_NSGA2
            Function to use to select individuals for survival. Must take a matrix of scores and return selected indexes.
            Used to selected population_size individuals at the start of each generation to use for mutation and crossover.

        parent_selector : function, default=parent_select_NSGA2
            Function to use to select pairs parents for crossover and individuals for mutation. Must take a matrix of scores and return selected indexes.

        budget_range : list [start, end], default=None
            A starting and ending budget to use for the budget scaling.

        budget_scaling float : [0,1], default=0.5
            A scaling factor to use when determining how fast we move the budget from the start to end budget.

        individuals_until_end_budget : int, default=1
            The number of generations to run before reaching the max budget.

        stepwise_steps : int, default=1
            The number of staircase steps to take when scaling the budget and population size.

        threshold_evaluation_early_stop : list [start, end], default=None
            starting and ending percentile to use as a threshold for the evaluation early stopping.
            Values between 0 and 100.

        threshold_evaluation_scaling : float [0,inf), default=0.5
            A scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.
            Must be greater than zero. Higher numbers will move the threshold to the end faster.

        min_history_threshold : int, default=0
            The minimum number of previous scores needed before using threshold early stopping.

        selection_evaluation_early_stop : list, default=None
            A lower and upper percent of the population size to select each round of CV.
            Values between 0 and 1.

        selection_evaluation_scaling : float, default=0.5
            A scaling factor to use when determining how fast we move the threshold moves from the start to end percentile.
            Must be greater than zero. Higher numbers will move the threshold to the end faster.

        n_initial_optimizations : int, default=0
            Number of individuals to optimize before starting the evolution.

        optimization_cv : int
           Number of folds to use for the optuna optimization's internal cross-validation.

        max_optimize_time_seconds : float, default=60*5
            Maximum time to run an optimization

        optimization_steps : int, default=10
            Number of steps per optimization

        warm_start : bool, default=False
            If True, will use the continue the evolutionary algorithm from the last generation of the previous run.

        subset_column : str or int, default=None
            EXPERIMENTAL The column to use for the subset selection. Must also pass in unique_subset_values to GraphIndividual to function.

        verbose : int, default=1
            How much information to print during the optimization process. Higher values include the information from lower values.
            0. nothing
            1. progress bar

            3. best individual
            4. warnings
            >=5. full warnings trace

        random_state : int, None, default=None
            A seed for reproducability of experiments. This value will be passed to numpy.random.default_rng() to create an instnce of the genrator to pass to other classes
            - int
                Will be used to create and lock in Generator instance with 'numpy.random.default_rng()'
            - None
                Will be used to create Generator for 'numpy.random.default_rng()' where a fresh, unpredictable entropy will be pulled from the OS


        periodic_checkpoint_folder : str, default=None
            Folder to save the population to periodically. If None, no periodic saving will be done.
            If provided, training will resume from this checkpoint.

        callback : tpot2.CallBackInterface, default=None
            Callback object. Not implemented

        processes : bool, default=True
            If True, will use multiprocessing to parallelize the optimization process. If False, will use threading.
            True seems to perform better. However, False is required for interactive debugging.

        Attributes
        ----------

        fitted_pipeline_ : GraphPipeline
            A fitted instance of the GraphPipeline that inherits from sklearn BaseEstimator. This is fitted on the full X, y passed to fit.

        evaluated_individuals : A pandas data frame containing data for all evaluated individuals in the run.
            Columns:
            - *objective functions : The first few columns correspond to the passed in scorers and objective functions
            - Parents : A tuple containing the indexes of the pipelines used to generate the pipeline of that row. If NaN, this pipeline was generated randomly in the initial population.
            - Variation_Function : Which variation function was used to mutate or crossover the parents. If NaN, this pipeline was generated randomly in the initial population.
            - Individual : The internal representation of the individual that is used during the evolutionary algorithm. This is not an sklearn BaseEstimator.
            - Generation : The generation the pipeline first appeared.
            - Pareto_Front	: The nondominated front that this pipeline belongs to. 0 means that its scores is not strictly dominated by any other individual.
                            To save on computational time, the best frontier is updated iteratively each generation.
                            The pipelines with the 0th pareto front do represent the exact best frontier. However, the pipelines with pareto front >= 1 are only in reference to the other pipelines in the final population.
                            All other pipelines are set to NaN.
            - Instance	: The unfitted GraphPipeline BaseEstimator.
            - *validation objective functions : Objective function scores evaluated on the validation set.
            - Validation_Pareto_Front : The full pareto front calculated on the validation set. This is calculated for all pipelines with Pareto_Front equal to 0. Unlike the Pareto_Front which only calculates the frontier and the final population, the Validation Pareto Front is calculated for all pipelines tested on the validation set.

        pareto_front : The same pandas dataframe as evaluated individuals, but containing only the frontier pareto front pipelines.
        '''

        # sklearn BaseEstimator must have a corresponding attribute for each parameter.
        # These should not be modified once set.

        self.scorers = scorers
        self.scorers_weights = scorers_weights
        self.classification = classification
        self.cv = cv
        self.other_objective_functions = other_objective_functions
        self.other_objective_functions_weights = other_objective_functions_weights
        self.objective_function_names = objective_function_names
        self.bigger_is_better = bigger_is_better
        self.max_size = max_size
        self.linear_pipeline = linear_pipeline
        self.root_config_dict= root_config_dict
        self.inner_config_dict= inner_config_dict
        self.leaf_config_dict= leaf_config_dict
        self.cross_val_predict_cv = cross_val_predict_cv
        self.categorical_features = categorical_features
        self.subsets = subsets
        self.memory = memory
        self.preprocessing = preprocessing
        self.validation_strategy = validation_strategy
        self.validation_fraction = validation_fraction
        self.disable_label_encoder = disable_label_encoder
        self.population_size = population_size
        self.initial_population_size = initial_population_size

        self.early_stop = early_stop
        self.early_stop_seconds = early_stop_seconds
        self.scorers_early_stop_tol = scorers_early_stop_tol
        self.other_objectives_early_stop_tol = other_objectives_early_stop_tol
        self.max_time_seconds = max_time_seconds
        self.max_eval_time_seconds = max_eval_time_seconds
        self.n_jobs= n_jobs
        self.memory_limit = memory_limit
        self.client = client

        self.crossover_probability = crossover_probability
        self.mutate_probability = mutate_probability
        self.mutate_then_crossover_probability= mutate_then_crossover_probability
        self.crossover_then_mutate_probability= crossover_then_mutate_probability
        self.survival_selector=survival_selector
        self.parent_selector=parent_selector
        self.budget_range = budget_range
        self.budget_scaling = budget_scaling
        self.individuals_until_end_budget = individuals_until_end_budget
        self.stepwise_steps = stepwise_steps

        self.warm_start = warm_start
        self.subset_column = subset_column

        self.verbose = verbose
        self.periodic_checkpoint_folder = periodic_checkpoint_folder
        self.callback = callback
        self.processes = processes


        self.scatter = scatter

        self.optuna_optimize_pareto_front = optuna_optimize_pareto_front
        self.optuna_optimize_pareto_front_trials = optuna_optimize_pareto_front_trials
        self.optuna_optimize_pareto_front_timeout = optuna_optimize_pareto_front_timeout
        self.optuna_storage = optuna_storage

        # create random number generator based on rng_seed
        self.rng = np.random.default_rng(random_state)
        # save random state passed to us for other functions that use random_state
        self.random_state = random_state
        # set the numpy seed so anything using it will be consistent as well
        np.random.seed(random_state)


        self.max_evaluated_individuals = max_evaluated_individuals

        #Initialize other used params


        if self.initial_population_size is None:
            self._initial_population_size = self.population_size
        else:
            self._initial_population_size = self.initial_population_size

        if isinstance(self.scorers, str):
            self._scorers = [self.scorers]

        elif callable(self.scorers):
            self._scorers = [self.scorers]
        else:
            self._scorers = self.scorers

        self._scorers = [sklearn.metrics.get_scorer(scoring) for scoring in self._scorers]
        self._scorers_early_stop_tol = self.scorers_early_stop_tol

        self._evolver = tpot2.evolvers.SteadyStateEvolver



        self.objective_function_weights = [*scorers_weights, *other_objective_functions_weights]


        if self.objective_function_names is None:
            obj_names = [f.__name__ for f in other_objective_functions]
        else:
            obj_names = self.objective_function_names
        self.objective_names = [f._score_func.__name__ if hasattr(f,"_score_func") else f.__name__ for f in self._scorers] + obj_names


        if not isinstance(self.other_objectives_early_stop_tol, list):
            self._other_objectives_early_stop_tol = [self.other_objectives_early_stop_tol for _ in range(len(self.other_objective_functions))]
        else:
            self._other_objectives_early_stop_tol = self.other_objectives_early_stop_tol

        if not isinstance(self._scorers_early_stop_tol, list):
            self._scorers_early_stop_tol = [self._scorers_early_stop_tol for _ in range(len(self._scorers))]
        else:
            self._scorers_early_stop_tol = self._scorers_early_stop_tol

        self.early_stop_tol = [*self._scorers_early_stop_tol, *self._other_objectives_early_stop_tol]

        self._evolver_instance = None
        self.evaluated_individuals = None

        self.label_encoder_ = None

        set_dask_settings()


    def fit(self, X, y):
        if self.client is not None: #If user passed in a client manually
           _client = self.client
        else:

            if self.verbose >= 4:
                silence_logs = 30
            elif self.verbose >=5:
                silence_logs = 40
            else:
                silence_logs = 50
            cluster = LocalCluster(n_workers=self.n_jobs, #if no client is passed in and no global client exists, create our own
                    threads_per_worker=1,
                    processes=self.processes,
                    silence_logs=silence_logs,
                    memory_limit=self.memory_limit)
            _client = Client(cluster)


        if self.classification and not self.disable_label_encoder and not check_if_y_is_encoded(y):
            warnings.warn("Labels are not encoded as ints from 0 to N. For compatibility with some classifiers such as sklearn, TPOT has encoded y with the sklearn LabelEncoder. When using pipelines outside the main TPOT estimator class, you can encode the labels with est.label_encoder_")
            self.label_encoder_ = LabelEncoder()
            y = self.label_encoder_.fit_transform(y)

        self.evaluated_individuals = None
        #determine validation strategy
        if self.validation_strategy == 'auto':
            nrows = X.shape[0]
            ncols = X.shape[1]

            if nrows/ncols < 20:
                validation_strategy = 'reshuffled'
            elif nrows/ncols < 100:
                validation_strategy = 'split'
            else:
                validation_strategy = 'none'
        else:
            validation_strategy = self.validation_strategy

        if validation_strategy == 'split':
            if self.classification:
                X, X_val, y, y_val = train_test_split(X, y, test_size=self.validation_fraction, stratify=y, random_state=self.random_state)
            else:
                X, X_val, y, y_val = train_test_split(X, y, test_size=self.validation_fraction, random_state=self.random_state)


        X_original = X
        y_original = y
        if isinstance(self.cv, int) or isinstance(self.cv, float):
            n_folds = self.cv
        else:
            n_folds = self.cv.get_n_splits(X, y)

        if self.classification:
            X, y = remove_underrepresented_classes(X, y, n_folds)

        if self.preprocessing:
            #X = pd.DataFrame(X)

            if not isinstance(self.preprocessing, bool) and isinstance(self.preprocessing, sklearn.base.BaseEstimator):
                self._preprocessing_pipeline = self.preprocessing

            #TODO: check if there are missing values in X before imputation. If not, don't include imputation in pipeline. Check if there are categorical columns. If not, don't include one hot encoding in pipeline
            else: #if self.preprocessing is True or not a sklearn estimator
                
                pipeline_steps = []

                if self.categorical_features is not None: #if categorical features are specified, use those
                    pipeline_steps.append(("impute_categorical", tpot2.builtin_modules.ColumnSimpleImputer(self.categorical_features, strategy='most_frequent')))
                    pipeline_steps.append(("impute_numeric", tpot2.builtin_modules.ColumnSimpleImputer("numeric", strategy='mean')))
                    pipeline_steps.append(("impute_categorical", tpot2.builtin_modules.ColumnOneHotEncoder(self.categorical_features, strategy='most_frequent')))

                else:
                    if isinstance(X, pd.DataFrame):
                        categorical_columns = X.select_dtypes(include=['object']).columns
                        if len(categorical_columns) > 0:
                            pipeline_steps.append(("impute_categorical", tpot2.builtin_modules.ColumnSimpleImputer("categorical", strategy='most_frequent')))
                            pipeline_steps.append(("impute_numeric", tpot2.builtin_modules.ColumnSimpleImputer("numeric", strategy='mean')))
                            pipeline_steps.append(("impute_categorical", tpot2.builtin_modules.ColumnOneHotEncoder("categorical", strategy='most_frequent')))
                        else:
                            pipeline_steps.append(("impute_numeric", tpot2.builtin_modules.ColumnSimpleImputer("all", strategy='mean')))
                    else:
                        pipeline_steps.append(("impute_numeric", tpot2.builtin_modules.ColumnSimpleImputer("all", strategy='mean')))
                            
                self._preprocessing_pipeline = sklearn.pipeline.Pipeline(pipeline_steps)

            X = self._preprocessing_pipeline.fit_transform(X, y)
            
        else:
            self._preprocessing_pipeline = None

        #_, y = sklearn.utils.check_X_y(X, y, y_numeric=True)

        #Set up the configuation dictionaries and the search spaces

        #check if self.cv is a number
        if isinstance(self.cv, int) or isinstance(self.cv, float):
            if self.classification:
                self.cv_gen = sklearn.model_selection.StratifiedKFold(n_splits=self.cv, shuffle=True, random_state=self.random_state)
            else:
                self.cv_gen = sklearn.model_selection.KFold(n_splits=self.cv, shuffle=True, random_state=self.random_state)

        else:
            self.cv_gen = sklearn.model_selection.check_cv(self.cv, y, classifier=self.classification)


        n_samples= int(math.floor(X.shape[0]/n_folds))
        n_features=X.shape[1]

        if isinstance(X, pd.DataFrame):
            self.feature_names = X.columns
        else:
            self.feature_names = None

        if self.root_config_dict == 'Auto':
            if self.classification:
                n_classes = len(np.unique(y))
                root_config_dict = get_configuration_dictionary("classifiers", n_samples, n_features, self.classification, self.random_state, self.cv_gen, subsets=self.subsets, feature_names=self.feature_names, n_classes=n_classes)
            else:
                root_config_dict = get_configuration_dictionary("regressors", n_samples, n_features, self.classification, self.random_state, self.cv_gen, subsets=self.subsets, feature_names=self.feature_names)
        else:
            root_config_dict = get_configuration_dictionary(self.root_config_dict, n_samples, n_features, self.classification, self.random_state, self.cv_gen, subsets=self.subsets,feature_names=self.feature_names)

        inner_config_dict = get_configuration_dictionary(self.inner_config_dict, n_samples, n_features, self.classification, self.random_state, self.cv_gen, subsets=self.subsets, feature_names=self.feature_names)
        leaf_config_dict = get_configuration_dictionary(self.leaf_config_dict, n_samples, n_features, self.classification, self.random_state, self.cv_gen, subsets=self.subsets, feature_names=self.feature_names)



        def objective_function(pipeline_individual,
                                            X,
                                            y,
                                            is_classification=self.classification,
                                            scorers= self._scorers,
                                            cv=self.cv_gen,
                                            other_objective_functions=self.other_objective_functions,
                                            memory=self.memory,
                                            cross_val_predict_cv=self.cross_val_predict_cv,
                                            subset_column=self.subset_column,
                                            **kwargs):
            return objective_function_generator(
                pipeline_individual,
                X,
                y,
                is_classification=is_classification,
                scorers= scorers,
                cv=cv,
                other_objective_functions=other_objective_functions,
                memory=memory,
                cross_val_predict_cv=cross_val_predict_cv,
                subset_column=subset_column,
                **kwargs,
            )

        self.individual_generator_instance = tpot2.individual_representations.graph_pipeline_individual.estimator_graph_individual_generator(
                                                            inner_config_dict=inner_config_dict,
                                                            root_config_dict=root_config_dict,
                                                            leaf_config_dict=leaf_config_dict,
                                                            max_size = self.max_size,
                                                            linear_pipeline=self.linear_pipeline,
                                                                )



        if self.scatter:
            X_future = _client.scatter(X)
            y_future = _client.scatter(y)
        else:
            X_future = X
            y_future = y

        #If warm start and we have an evolver instance, use the existing one
        if not(self.warm_start and self._evolver_instance is not None):
            self._evolver_instance = self._evolver(   individual_generator=self.individual_generator_instance,
                                            objective_functions= [objective_function],
                                            objective_function_weights = self.objective_function_weights,
                                            objective_names=self.objective_names,
                                            bigger_is_better = self.bigger_is_better,
                                            population_size= self.population_size,

                                            initial_population_size = self._initial_population_size,
                                            n_jobs=self.n_jobs,
                                            verbose = self.verbose,
                                            max_time_seconds =      self.max_time_seconds ,
                                            max_eval_time_seconds = self.max_eval_time_seconds,



                                            periodic_checkpoint_folder = self.periodic_checkpoint_folder,


                                            early_stop_tol = self.early_stop_tol,
                                            early_stop= self.early_stop,
                                            early_stop_seconds =  self.early_stop_seconds,

                                            budget_range = self.budget_range,
                                            budget_scaling = self.budget_scaling,
                                            individuals_until_end_budget = self.individuals_until_end_budget,


                                            stepwise_steps = self.stepwise_steps,
                                            client = _client,
                                            objective_kwargs = {"X": X_future, "y": y_future},
                                            survival_selector=self.survival_selector,
                                            parent_selector=self.parent_selector,

                                            crossover_probability = self.crossover_probability,
                                            mutate_probability = self.mutate_probability,
                                            mutate_then_crossover_probability= self.mutate_then_crossover_probability,
                                            crossover_then_mutate_probability= self.crossover_then_mutate_probability,


                                            max_evaluated_individuals = self.max_evaluated_individuals,

                                            rng_=self.rng,
                                            )


        self._evolver_instance.optimize()
        #self._evolver_instance.population.update_pareto_fronts(self.objective_names, self.objective_function_weights)
        self.make_evaluated_individuals()


        if self.optuna_optimize_pareto_front:
            pareto_front_inds = self.pareto_front['Individual'].values
            all_graphs, all_scores = tpot2.individual_representations.graph_pipeline_individual.simple_parallel_optuna(pareto_front_inds,  objective_function, self.objective_function_weights, _client, storage=self.optuna_storage, steps=self.optuna_optimize_pareto_front_trials, verbose=self.verbose, max_eval_time_seconds=self.max_eval_time_seconds, max_time_seconds=self.optuna_optimize_pareto_front_timeout, **{"X": X, "y": y})
            all_scores = tpot2.utils.eval_utils.process_scores(all_scores, len(self.objective_function_weights))

            if len(all_graphs) > 0:
                df = pd.DataFrame(np.column_stack((all_graphs, all_scores,np.repeat("Optuna",len(all_graphs)))), columns=["Individual"] + self.objective_names +["Parents"])
                for obj in self.objective_names:
                    df[obj] = df[obj].apply(convert_to_float)

                self.evaluated_individuals = pd.concat([self.evaluated_individuals, df], ignore_index=True)
            else:
                print("WARNING NO OPTUNA TRIALS COMPLETED")

        tpot2.utils.get_pareto_frontier(self.evaluated_individuals, column_names=self.objective_names, weights=self.objective_function_weights, invalid_values=["TIMEOUT","INVALID"])

        if validation_strategy == 'reshuffled':
            best_pareto_front_idx = list(self.pareto_front.index)
            best_pareto_front = list(self.pareto_front.loc[best_pareto_front_idx]['Individual'])

            #reshuffle rows
            X, y = sklearn.utils.shuffle(X, y, random_state=self.random_state)

            if self.scatter:
                X_future = _client.scatter(X)
                y_future = _client.scatter(y)
            else:
                X_future = X
                y_future = y

            val_objective_function_list = [lambda   ind,
                                                    X,
                                                    y,
                                                    is_classification=self.classification,
                                                    scorers= self._scorers,
                                                    cv=self.cv_gen,
                                                    other_objective_functions=self.other_objective_functions,
                                                    memory=self.memory,
                                                    cross_val_predict_cv=self.cross_val_predict_cv,
                                                    subset_column=self.subset_column,
                                                    **kwargs: objective_function_generator(
                                                                                                ind,
                                                                                                X,
                                                                                                y,
                                                                                                is_classification=is_classification,
                                                                                                scorers= scorers,
                                                                                                cv=cv,
                                                                                                other_objective_functions=other_objective_functions,
                                                                                                memory=memory,
                                                                                                cross_val_predict_cv=cross_val_predict_cv,
                                                                                                subset_column=subset_column,
                                                                                                **kwargs,
                                                                                                )]

            objective_kwargs = {"X": X_future, "y": y_future}
            val_scores = tpot2.utils.eval_utils.parallel_eval_objective_list(
                best_pareto_front,
                val_objective_function_list, n_jobs=self.n_jobs, verbose=self.verbose, timeout=self.max_eval_time_seconds,n_expected_columns=len(self.objective_names), client=_client, **objective_kwargs)

            val_objective_names = ['validation_'+name for name in self.objective_names]
            self.objective_names_for_selection = val_objective_names
            self.evaluated_individuals.loc[best_pareto_front_idx,val_objective_names] = val_scores

            self.evaluated_individuals["Validation_Pareto_Front"] = tpot2.utils.get_pareto_front(self.evaluated_individuals, val_objective_names, self.objective_function_weights, invalid_values=["TIMEOUT","INVALID"])

        elif validation_strategy == 'split':


            if self.scatter:
                X_future = _client.scatter(X)
                y_future = _client.scatter(y)
                X_val_future = _client.scatter(X_val)
                y_val_future = _client.scatter(y_val)
            else:
                X_future = X
                y_future = y
                X_val_future = X_val
                y_val_future = y_val

            objective_kwargs = {"X": X_future, "y": y_future, "X_val" : X_val_future, "y_val":y_val_future }

            best_pareto_front_idx = list(self.pareto_front.index)
            best_pareto_front = list(self.pareto_front.loc[best_pareto_front_idx]['Individual'])
            val_objective_function_list = [lambda   ind,
                                                    X,
                                                    y,
                                                    X_val,
                                                    y_val,
                                                    scorers= self._scorers,
                                                    other_objective_functions=self.other_objective_functions,
                                                    memory=self.memory,
                                                    cross_val_predict_cv=self.cross_val_predict_cv,
                                                    subset_column=self.subset_column,
                                                    **kwargs: val_objective_function_generator(
                                                        ind,
                                                        X,
                                                        y,
                                                        X_val,
                                                        y_val,
                                                        scorers= scorers,
                                                        other_objective_functions=other_objective_functions,
                                                        memory=memory,
                                                        cross_val_predict_cv=cross_val_predict_cv,
                                                        subset_column=subset_column,
                                                        **kwargs,
                                                        )]

            val_scores = tpot2.utils.eval_utils.parallel_eval_objective_list(
                best_pareto_front,
                val_objective_function_list, n_jobs=self.n_jobs, verbose=self.verbose, timeout=self.max_eval_time_seconds,n_expected_columns=len(self.objective_names),client=_client, **objective_kwargs)

            val_objective_names = ['validation_'+name for name in self.objective_names]
            self.objective_names_for_selection = val_objective_names
            self.evaluated_individuals.loc[best_pareto_front_idx,val_objective_names] = val_scores
            self.evaluated_individuals["Validation_Pareto_Front"] = tpot2.utils.get_pareto_front(self.evaluated_individuals, val_objective_names, self.objective_function_weights, invalid_values=["TIMEOUT","INVALID"])
        else:
            self.objective_names_for_selection = self.objective_names

        val_scores = self.evaluated_individuals[~self.evaluated_individuals[self.objective_names_for_selection].isin(["TIMEOUT","INVALID"]).any(axis=1)][self.objective_names_for_selection].astype(float)
        weighted_scores = val_scores*self.objective_function_weights

        if self.bigger_is_better:
            best_idx = weighted_scores[self.objective_names_for_selection[0]].idxmax()
        else:
            best_idx = weighted_scores[self.objective_names_for_selection[0]].idxmin()

        best_individual = self.evaluated_individuals.loc[best_idx]['Individual']
        self.selected_best_score =  self.evaluated_individuals.loc[best_idx]


        best_individual_pipeline = best_individual.export_pipeline(memory=self.memory, cross_val_predict_cv=self.cross_val_predict_cv, subset_column=self.subset_column)

        if self.preprocessing:
            self.fitted_pipeline_ = sklearn.pipeline.make_pipeline(sklearn.base.clone(self._preprocessing_pipeline), best_individual_pipeline )
        else:
            self.fitted_pipeline_ = best_individual_pipeline

        self.fitted_pipeline_.fit(X_original,y_original) #TODO use y_original as well?


        if self.client is None: #no client was passed in
            #close cluster and client
            _client.close()
            cluster.close()

        return self

    def _estimator_has(attr):
        '''Check if we can delegate a method to the underlying estimator.
        First, we check the first fitted final estimator if available, otherwise we
        check the unfitted final estimator.
        '''
        return  lambda self: (self.fitted_pipeline_ is not None and
            hasattr(self.fitted_pipeline_, attr)
        )






    @available_if(_estimator_has('predict'))
    def predict(self, X, **predict_params):
        check_is_fitted(self)
        #X = check_array(X)
        preds = self.fitted_pipeline_.predict(X,**predict_params)
        if self.classification and self.label_encoder_:
            preds = self.label_encoder_.inverse_transform(preds)

        return preds

    @available_if(_estimator_has('predict_proba'))
    def predict_proba(self, X, **predict_params):
        check_is_fitted(self)
        #X = check_array(X)
        return self.fitted_pipeline_.predict_proba(X,**predict_params)

    @available_if(_estimator_has('decision_function'))
    def decision_function(self, X, **predict_params):
        check_is_fitted(self)
        #X = check_array(X)
        return self.fitted_pipeline_.decision_function(X,**predict_params)

    @available_if(_estimator_has('transform'))
    def transform(self, X, **predict_params):
        check_is_fitted(self)
        #X = check_array(X)
        return self.fitted_pipeline_.transform(X,**predict_params)

    @property
    def classes_(self):
        """The classes labels. Only exist if the last step is a classifier."""

        if self.label_encoder_:
            return self.label_encoder_.classes_
        else:
            return self.fitted_pipeline_.classes_

    @property
    def _estimator_type(self):
        return self.fitted_pipeline_._estimator_type

    def make_evaluated_individuals(self):
        #check if _evolver_instance exists
        if self.evaluated_individuals is None:
            self.evaluated_individuals  =  self._evolver_instance.population.evaluated_individuals.copy()
            objects = list(self.evaluated_individuals.index)
            object_to_int = dict(zip(objects, range(len(objects))))
            self.evaluated_individuals = self.evaluated_individuals.set_index(self.evaluated_individuals.index.map(object_to_int))
            self.evaluated_individuals['Parents'] = self.evaluated_individuals['Parents'].apply(lambda row: convert_parents_tuples_to_integers(row, object_to_int))

            self.evaluated_individuals["Instance"] = self.evaluated_individuals["Individual"].apply(lambda ind: apply_make_pipeline(ind, preprocessing_pipeline=self._preprocessing_pipeline))

        return self.evaluated_individuals

    @property
    def pareto_front(self):
        #check if _evolver_instance exists
        if self.evaluated_individuals is None:
            return None
        else:
            if "Pareto_Front" not in self.evaluated_individuals:
                return self.evaluated_individuals
            else:
                return self.evaluated_individuals[self.evaluated_individuals["Pareto_Front"]==1]




#TODO: are all the imports in the init files done correctly?
#TODO clean up import organization

from .graphsklearn import GraphPipeline
from .population import Population

from . import builtin_modules
from . import utils
from . import config
from . import individual_representations
from . import evolvers
from . import objectives
from . import selectors
from . import tpot_estimator


from .tpot_estimator import TPOTClassifier, TPOTRegressor, TPOTEstimator, TPOTEstimatorSteadyState



from numpy import iterable
import tpot2
import numpy as np
from .. import BaseIndividual

class SubsetSelector(BaseIndividual):
    def __init__(   self,
                    values,
                    rng_=None,
                    initial_set = None,
                    k=1, #step size for shuffling
                ):

        rng = np.random.default_rng(rng_)

        if isinstance(values, int):
            self.values = set(range(0,values))
        else:
            self.values = set(values)


        if initial_set is None:
            self.subsets = set(rng.choices(values, k=k))
        else:
            self.subsets = set(initial_set)

        self.k = k

        self.mutation_list = [self._mutate_add, self._mutate_remove]
        self.crossover_list = [self._crossover_swap]

    def _mutate_add(self, rng_=None):
        rng = np.random.default_rng(rng_)
        not_included = list(self.values.difference(self.subsets))
        if len(not_included) > 1:
            self.subsets.update(rng.choice(not_included, k=min(self.k, len(not_included))))
            return True
        else:
            return False

    def _mutate_remove(self, rng_=None):
        rng = np.random.default_rng(rng_)
        if len(self.subsets) > 1:
            self.subsets = self.subsets - set(rng.choice(list(self.subsets), k=min(self.k, len(self.subsets)-1) ))

    def _crossover_swap(self, ss2, rng_=None):
        rng = np.random.default_rng(rng_)
        diffs = self.subsets.symmetric_difference(ss2.subsets)

        if len(diffs) == 0:
            return False
        for v in diffs:
            self.subsets.discard(v)
            ss2.subsets.discard(v)
            rng.choice([self.subsets, ss2.subsets]).add(v)

        return True



from sklearn.linear_model import SGDRegressor
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from sklearn.linear_model import Lasso
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import Lars
from sklearn.linear_model import LassoLars, LassoLarsCV
from sklearn.linear_model import RidgeCV


from sklearn.svm import SVR
from sklearn.svm import LinearSVR

from sklearn.ensemble import AdaBoostRegressor, GradientBoostingRegressor,RandomForestRegressor
from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.linear_model import ElasticNetCV

from xgboost import XGBRegressor
from functools import partial




#TODO: fill in remaining
#TODO check for places were we could use log scaling

def params_RandomForestRegressor(trial, random_state=None, name=None):
    return {
        'n_estimators': 100,
        'max_features': trial.suggest_float(f'max_features_{name}', 0.05, 1.0),
        'bootstrap': trial.suggest_categorical(name=f'bootstrap_{name}', choices=[True, False]),
        'min_samples_split': trial.suggest_int(f'min_samples_split_{name}', 2, 21),
        'min_samples_leaf': trial.suggest_int(f'min_samples_leaf_{name}', 1, 21),
        'random_state': random_state
    }


# SGDRegressor parameters
def params_SGDRegressor(trial, random_state=None, name=None):
    params = {
        'loss': trial.suggest_categorical(f'loss_{name}', ['huber', 'squared_error', 'epsilon_insensitive', 'squared_epsilon_insensitive']),
        'penalty': 'elasticnet',
        'alpha': trial.suggest_float(f'alpha_{name}', 1e-5, 0.01, log=True),
        'learning_rate': trial.suggest_categorical(f'learning_rate_{name}', ['invscaling', 'constant']),
        'fit_intercept':True,
        'l1_ratio': trial.suggest_float(f'l1_ratio_{name}', 0.0, 1.0),
        'eta0': trial.suggest_float(f'eta0_{name}', 0.01, 1.0),
        'power_t': trial.suggest_float(f'power_t_{name}', 1e-5, 100.0, log=True),
        'random_state': random_state

    }
    return params

# Ridge parameters
def params_Ridge(trial, random_state=None, name=None):
    params = {
        'alpha': trial.suggest_float(f'alpha_{name}', 0.0, 1.0),
        'fit_intercept': True,


        #'max_iter': trial.suggest_int(f'max_iter_{name}', 100, 1000),
        'tol': trial.suggest_float(f'tol_{name}', 1e-5, 1e-1, log=True),
        'solver': trial.suggest_categorical(f'solver_{name}', ['auto', 'svd', 'cholesky', 'lsqr', 'sparse_cg', 'sag', 'saga']),
        'random_state': random_state
    }
    return params


# Lasso parameters
def params_Lasso(trial, random_state=None, name=None):
    params = {
        'alpha': trial.suggest_float(f'alpha_{name}', 0.0, 1.0),
        'fit_intercept': True,
        # 'normalize': trial.suggest_categorical(f'normalize_{name}', [True, False]),
        'precompute': trial.suggest_categorical(f'precompute_{name}', [True, False, 'auto']),

        #'max_iter': trial.suggest_int(f'max_iter_{name}', 100, 1000),
        'tol': trial.suggest_float(f'tol_{name}', 1e-5, 1e-1, log=True),

        'positive': trial.suggest_categorical(f'positive_{name}', [True, False]),
        'selection': trial.suggest_categorical(f'selection_{name}', ['cyclic', 'random']),
        'random_state': random_state
    }
    return params

# ElasticNet parameters
def params_ElasticNet(trial, random_state=None, name=None):
    params = {
        'alpha': 1 - trial.suggest_float(f'alpha_{name}', 0.0, 1.0, log=True),
        'l1_ratio': 1- trial.suggest_float(f'l1_ratio_{name}',0.0, 1.0),
        'random_state': random_state
        }
    return params

# Lars parameters
def params_Lars(trial, random_state=None, name=None):
    params = {
        'fit_intercept': True,
        'verbose': trial.suggest_categorical(f'verbose_{name}', [True, False]),
        'normalize': trial.suggest_categorical(f'normalize_{name}', [True, False]),

        # 'precompute': trial.suggest_categorical(f'precompute_{name}', ['auto_{name}', True, False]),
        'n_nonzero_coefs': trial.suggest_int(f'n_nonzero_coefs_{name}', 1, 100),
        'eps': trial.suggest_float(f'eps_{name}', 1e-5, 1e-1, log=True),
        'copy_X': trial.suggest_categorical(f'copy_X_{name}', [True, False]),
        'fit_path': trial.suggest_categorical(f'fit_path_{name}', [True, False]),
        # 'positive': trial.suggest_categorical(f'positive_{name}', [True, False]),
        'random_state': random_state
    }
    return params

# OrthogonalMatchingPursuit parameters
def params_OrthogonalMatchingPursuit(trial, name=None):
    params = {
        'n_nonzero_coefs': trial.suggest_int(f'n_nonzero_coefs_{name}', 1, 100),
        'tol': trial.suggest_float(f'tol_{name}', 1e-5, 1e-1, log=True),
        'fit_intercept': True,
        'normalize': trial.suggest_categorical(f'normalize_{name}', [True, False]),
        'precompute': trial.suggest_categorical(f'precompute_{name}', ['auto', True, False]),
    }
    return params

# BayesianRidge parameters
def params_BayesianRidge(trial, name=None):
    params = {
        'n_iter': trial.suggest_int(f'n_iter_{name}', 100, 1000),
        'tol': trial.suggest_float(f'tol_{name}', 1e-5, 1e-1, log=True),
        'alpha_1': trial.suggest_float(f'alpha_1_{name}', 1e-6, 1e-1, log=True),
        'alpha_2': trial.suggest_float(f'alpha_2_{name}', 1e-6, 1e-1, log=True),
        'lambda_1': trial.suggest_float(f'lambda_1_{name}', 1e-6, 1e-1, log=True),
        'lambda_2': trial.suggest_float(f'lambda_2_{name}', 1e-6, 1e-1, log=True),
        'compute_score': trial.suggest_categorical(f'compute_score_{name}', [True, False]),
        'fit_intercept': True,
        'normalize': trial.suggest_categorical(f'normalize_{name}', [True, False]),
        'copy_X': trial.suggest_categorical(f'copy_X_{name}', [True, False]),
    }
    return params

# LassoLars parameters
def params_LassoLars(trial, random_state=None, name=None):
    params = {
        'alpha': trial.suggest_float(f'alpha_{name}', 0.0, 1.0),
        # 'fit_intercept': True,
        # 'normalize': trial.suggest_categorical(f'normalize_{name}', [True, False]),
        # 'precompute': trial.suggest_categorical(f'precompute_{name}', ['auto_{name}', True, False]),
        #'max_iter': trial.suggest_int(f'max_iter_{name}', 100, 1000),
        'eps': trial.suggest_float(f'eps_{name}', 1e-5, 1e-1, log=True),
        # 'copy_X': trial.suggest_categorical(f'copy_X_{name}', [True, False]),
        # 'positive': trial.suggest_categorical(f'positive_{name}', [True, False]),
        'random_state': random_state
    }
    return params

# LassoLars parameters
def params_LassoLarsCV(trial, cv, name=None):
    params = {
        'normalize': trial.suggest_categorical(f'normalize_{name}', [True, False]),
        'cv': cv,
    }
    return params

# BaggingRegressor parameters
def params_BaggingRegressor(trial, random_state=None, name=None):
    params = {
        'n_estimators': trial.suggest_int(f'n_estimators_{name}', 10, 100),
        'max_samples': trial.suggest_float(f'max_samples_{name}', 0.05, 1.00),
        'max_features': trial.suggest_float(f'max_features_{name}', 0.05, 1.00),
        'bootstrap': trial.suggest_categorical(f'bootstrap_{name}', [True, False]),
        'bootstrap_features': trial.suggest_categorical(f'bootstrap_features_{name}', [True, False]),
        'random_state': random_state
    }
    return params

# ARDRegression parameters
def params_ARDRegression(trial, name=None):
    params = {
        'n_iter': trial.suggest_int(f'n_iter_{name}', 100, 1000),
        'tol': trial.suggest_float(f'tol_{name}', 1e-5, 1e-1, log=True),
        'alpha_1': trial.suggest_float(f'alpha_1_{name}', 1e-6, 1e-1, log=True),
        'alpha_2': trial.suggest_float(f'alpha_2_{name}', 1e-6, 1e-1, log=True),
        'lambda_1': trial.suggest_float(f'lambda_1_{name}', 1e-6, 1e-1, log=True),
        'lambda_2': trial.suggest_float(f'lambda_2_{name}', 1e-6, 1e-1, log=True),
        'compute_score': trial.suggest_categorical(f'compute_score_{name}', [True, False]),
        'threshold_lambda': trial.suggest_int(f'threshold_lambda_{name}', 100, 1000),
        'fit_intercept': True,
        'normalize': trial.suggest_categorical(f'normalize_{name}', [True, False]),
        'copy_X': trial.suggest_categorical(f'copy_X_{name}', [True, False]),
    }
    return params



# TheilSenRegressor parameters
def params_TheilSenRegressor(trial, random_state=None, name=None):
    params = {
        'n_subsamples': trial.suggest_int(f'n_subsamples_{name}', 10, 100),
        'max_subpopulation': trial.suggest_int(f'max_subpopulation_{name}', 100, 1000),
        'fit_intercept': True,
        'copy_X': trial.suggest_categorical(f'copy_X_{name}', [True, False]),
        'verbose': trial.suggest_categorical(f'verbose_{name}', [True, False]),
        'random_state': random_state
    }
    return params


# SVR parameters
def params_SVR(trial, name=None):
    params = {
        'kernel': trial.suggest_categorical(name=f'kernel_{name}', choices=['poly', 'rbf', 'linear', 'sigmoid']),
        'C': trial.suggest_float(f'C_{name}', 1e-4, 25, log=True),
        'degree': trial.suggest_int(f'degree_{name}', 1, 4),
        'max_iter': 3000,
        'tol': 0.005,
    }
    return params

# Perceptron parameters
def params_Perceptron(trial, random_state=None, name=None):
    params = {
        'penalty': trial.suggest_categorical(f'penalty_{name}', [None, 'l2', 'l1', 'elasticnet']),
        'alpha': trial.suggest_float(f'alpha_{name}', 1e-5, 1e-1, log=True),
        'l1_ratio': trial.suggest_float(f'l1_ratio_{name}', 0.0, 1.0),
        'fit_intercept': True,
        #'max_iter': trial.suggest_int(f'max_iter_{name}', 100, 1000),
        'tol': trial.suggest_float(f'tol_{name}', 1e-5, 1e-1, log=True),
        'shuffle': trial.suggest_categorical(f'shuffle_{name}', [True, False]),
        'verbose': trial.suggest_categorical(f'verbose_{name}', [0, 1, 2, 3, 4, 5]),
        'eta0': trial.suggest_float(f'eta0_{name}', 1e-5, 1e-1, log=True),
        'learning_rate': trial.suggest_categorical(f'learning_rate_{name}', ['constant', 'optimal', 'invscaling']),
        'early_stopping': trial.suggest_categorical(f'early_stopping_{name}', [True, False]),
        'validation_fraction': trial.suggest_float(f'validation_fraction_{name}', 0.05, 1.00),
        'n_iter_no_change': trial.suggest_int(f'n_iter_no_change_{name}', 1, 100),
        'class_weight': trial.suggest_categorical(f'class_weight_{name}', [None, 'balanced']),
        'warm_start': trial.suggest_categorical(f'warm_start_{name}', [True, False]),
        'average': trial.suggest_categorical(f'average_{name}', [True, False]),
        'random_state': random_state
    }
    return params

def params_MLPRegressor(trial, random_state=None, name=None):
    params = {
        'alpha': trial.suggest_float(f'alpha_{name}', 1e-4, 1e-1, log=True),
        'learning_rate_init': trial.suggest_float(f'learning_rate_init_{name}', 1e-3, 1., log=True),
        'random_state': random_state
    }

    return params


#GradientBoostingRegressor parameters
def params_GradientBoostingRegressor(trial, random_state=None, name=None):
    loss = trial.suggest_categorical(f'loss_{name}', ['ls', 'lad', 'huber', 'quantile'])

    params = {

        'n_estimators': 100,
        'loss': loss,
        'learning_rate': trial.suggest_float(f'learning_rate_{name}', 1e-4, 1, log=True),
        'max_depth': trial.suggest_int(f'max_depth_{name}', 1, 11),
        'min_samples_split': trial.suggest_int(f'min_samples_split_{name}', 2, 21),
        'min_samples_leaf': trial.suggest_int(f'min_samples_leaf_{name}', 1, 21),
        'subsample': 1-trial.suggest_float(f'subsample_{name}', 0.05, 1.00, log=True),
        'max_features': 1-trial.suggest_float(f'max_features_{name}', 0.05, 1.00, log=True),
        'random_state': random_state

    }

    if loss == 'quantile' or loss == 'huber':
        alpha = trial.suggest_float(f'alpha_{name}', 0.05, 0.95)
        params['alpha'] = alpha

    return params



def params_DecisionTreeRegressor(trial, random_state=None, name=None):
    params = {
        'max_depth': trial.suggest_int(f'max_depth_{name}', 1,11),
        'min_samples_split': trial.suggest_int(f'min_samples_split_{name}', 2, 21),
        'min_samples_leaf': trial.suggest_int(f'min_samples_leaf_{name}', 1, 21),
        # 'criterion': trial.suggest_categorical(f'criterion_{name}', ['squared_error', 'friedman_mse', 'absolute_error', 'poisson']),
        # 'splitter': trial.suggest_categorical(f'splitter_{name}', ['best', 'random']),
        #'max_features': trial.suggest_categorical(f'max_features_{name}', [None, 'auto', 'sqrt', 'log2']),
        #'ccp_alpha': trial.suggest_float(f'ccp_alpha_{name}',  1e-1, 10.0),
        'random_state': random_state

    }
    return params

def params_KNeighborsRegressor(trial, name=None, n_samples=100):
    params = {
        'n_neighbors': trial.suggest_int(f'n_neighbors_{name}', 1, n_samples, log=True ),
        'weights': trial.suggest_categorical(f'weights_{name}', ['uniform', 'distance']),
        'p': trial.suggest_int(f'p_{name}', 1, 3),
        'metric': trial.suggest_categorical(f'metric_{name}', ['minkowski', 'euclidean', 'manhattan']),

        }
    return params

def params_LinearSVR(trial, random_state=None, name=None):
    params = {
        'epsilon': trial.suggest_float(f'epsilon_{name}', 1e-4, 1.0, log=True),
        'C': trial.suggest_float(f'C_{name}', 1e-4,25.0, log=True),
        'dual': trial.suggest_categorical(f'dual_{name}', [True,False]),
        'loss': trial.suggest_categorical(f'loss_{name}', ['epsilon_insensitive', 'squared_epsilon_insensitive']),
        'random_state': random_state

    }
    return params


# XGBRegressor parameters
def params_XGBRegressor(trial, random_state=None, name=None):
    return {
        'learning_rate': trial.suggest_float(f'learning_rate_{name}', 1e-3, 1, log=True),
        'subsample': trial.suggest_float(f'subsample_{name}', 0.05, 1.0),
        'min_child_weight': trial.suggest_int(f'min_child_weight_{name}', 1, 21),
        #'booster': trial.suggest_categorical(name='booster_{name}', choices=['gbtree', 'dart']),
        'n_estimators': 100,
        'max_depth': trial.suggest_int(f'max_depth_{name}', 1, 11),
        'nthread': 1,
        'verbosity': 0,
        'objective': 'reg:squarederror',
        'random_state': random_state
    }


def params_AdaBoostRegressor(trial, random_state=None, name=None):
    params = {
        'n_estimators': 100,
        'learning_rate': trial.suggest_float(f'learning_rate_{name}', 1e-3, 1.0, log=True),
        'loss': trial.suggest_categorical(f'loss_{name}', ['linear', 'square', 'exponential']),
        'random_state': random_state

    }
    return params

# ExtraTreesRegressor parameters
def params_ExtraTreesRegressor(trial, random_state=None, name=None):
    params = {
        'n_estimators': 100,
        'max_features': trial.suggest_float(f'max_features_{name}', 0.05, 1.0),
        'min_samples_split': trial.suggest_int(f'min_samples_split_{name}', 2, 21),
        'min_samples_leaf': trial.suggest_int(f'min_samples_leaf_{name}', 1, 21),
        'bootstrap': trial.suggest_categorical(f'bootstrap_{name}', [True, False]),

        #'criterion': trial.suggest_categorical(f'criterion_{name}', ['squared_error', 'poisson', 'absolute_error', 'friedman_mse']),

        #'max_depth': trial.suggest_int(f'max_depth_{name}', 1, 10),

        #'min_weight_fraction_leaf': trial.suggest_float(f'min_weight_fraction_leaf_{name}', 0.0, 0.5),
        # 'max_features': trial.suggest_categorical(f'max_features_{name}', [None, 'auto', 'sqrt', 'log2']),
        #'max_leaf_nodes': trial.suggest_int(f'max_leaf_nodes_{name}', 2, 100),
        #'min_impurity_decrease': trial.suggest_float(f'min_impurity_decrease_{name}', 1e-5, 1e-1, log=True),
        # 'min_impurity_split': trial.suggest_float(f'min_impurity_split_{name}', 1e-5, 1e-1, log=True),

        #if bootstrap is True
        #'oob_score': trial.suggest_categorical(f'oob_score_{name}', [True, False]),

        #'ccp_alpha': trial.suggest_float(f'ccp_alpha_{name}', 1e-5, 1e-1, log=True),
        # 'max_samples': trial.suggest_float(f'max_samples_{name}', 0.05, 1.00),

        'random_state': random_state
    }
    return params



def make_regressor_config_dictionary(random_state=None, cv=None, n_samples=10):
    n_samples = min(n_samples,100) #TODO optimize this


    regressor_config_dictionary = {
        #ElasticNet: params_ElasticNet,
        ElasticNetCV: {
                        'l1_ratio': [.1, .5, .7, .9, .95, .99, 1],
                        'cv': cv,
                        },
        ExtraTreesRegressor: partial(params_ExtraTreesRegressor, random_state=random_state),
        GradientBoostingRegressor: partial(params_GradientBoostingRegressor, random_state=random_state),
        AdaBoostRegressor: partial(params_AdaBoostRegressor, random_state=random_state),
        DecisionTreeRegressor: partial(params_DecisionTreeRegressor, random_state=random_state),
        KNeighborsRegressor: partial(params_KNeighborsRegressor,n_samples=n_samples),
        LassoLarsCV: partial(params_LassoLarsCV, cv=cv),
        SVR: params_SVR,
        RandomForestRegressor: partial(params_RandomForestRegressor, random_state=random_state),
        RidgeCV: {'cv': cv},
        XGBRegressor: partial(params_XGBRegressor, random_state=random_state),
        SGDRegressor: partial(params_SGDRegressor, random_state= random_state),

    }

    return regressor_config_dictionary


from sklearn.base import BaseEstimator, TransformerMixin

class Passthrough(TransformerMixin,BaseEstimator):

    def fit(self, X=None, y=None):
        return self

    def transform(self, X):
        return X



from tpot2 import GraphPipeline

from sklearn.linear_model import SGDClassifier
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier, GradientBoostingClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from lightgbm import LGBMClassifier
from sklearn.svm import LinearSVC

from functools import partial
#import GaussianNB

from sklearn.naive_bayes import GaussianNB, BernoulliNB, MultinomialNB

from sklearn.linear_model import SGDRegressor
from sklearn.linear_model import LinearRegression
from sklearn.linear_model import Ridge
from sklearn.linear_model import Lasso
from sklearn.linear_model import ElasticNet
from sklearn.linear_model import Lars
from sklearn.linear_model import LassoLars, LassoLarsCV
from sklearn.linear_model import RidgeCV


from sklearn.svm import SVR
from sklearn.svm import LinearSVR

from sklearn.ensemble import AdaBoostRegressor, GradientBoostingRegressor,RandomForestRegressor
from sklearn.ensemble import BaggingRegressor
from sklearn.ensemble import ExtraTreesRegressor
from sklearn.tree import DecisionTreeRegressor
from sklearn.neighbors import KNeighborsRegressor
from sklearn.linear_model import ElasticNetCV



from xgboost import XGBRegressor
from functools import partial


#TODO: how to best support transformers/selectors that take other transformers with their own hyperparameters? 
import numpy as np
from sklearn.feature_selection import SelectFwe
from sklearn.feature_selection import SelectPercentile
from sklearn.feature_selection import VarianceThreshold
from sklearn.feature_selection import RFE
from sklearn.feature_selection import SelectFromModel
import sklearn.feature_selection
from functools import partial
from sklearn.ensemble import ExtraTreesRegressor, ExtraTreesClassifier
from tpot2.builtin_modules import RFE_ExtraTreesClassifier, SelectFromModel_ExtraTreesClassifier, RFE_ExtraTreesRegressor, SelectFromModel_ExtraTreesRegressor


from functools import partial
from tpot2.builtin_modules import ZeroCount, OneHotEncoder
from sklearn.preprocessing import Binarizer
from sklearn.decomposition import FastICA
from sklearn.cluster import FeatureAgglomeration
from sklearn.preprocessing import MaxAbsScaler
from sklearn.preprocessing import MinMaxScaler
from sklearn.preprocessing import Normalizer
from sklearn.kernel_approximation import Nystroem
from sklearn.decomposition import PCA
from sklearn.preprocessing import PolynomialFeatures
from sklearn.kernel_approximation import RBFSampler
from sklearn.preprocessing import RobustScaler
from sklearn.preprocessing import StandardScaler

    # MultinomialNB: params_MultinomialNB,



#https://scikit-learn.org/stable/auto_examples/applications/plot_model_complexity_influence.html
def _count_nonzero_coefficients_and_intercept(est):
    n_coef = np.count_nonzero(est.coef_)
    if hasattr(est, 'intercept_'):
        n_coef += np.count_nonzero(est.intercept_)

    return n_coef

#https://stackoverflow.com/questions/51139875/sklearn-randomforestregressor-number-of-trainable-parameters
def tree_complexity(tree):
    return tree.tree_.node_count * 5 #each node has 5 parameters

#https://stackoverflow.com/questions/51139875/sklearn-randomforestregressor-number-of-trainable-parameters
def forest_complexity(forest):
    all_trees = np.array(forest.estimators_)
    if len(all_trees.shape)>1:
        all_trees = all_trees.ravel()
    return sum(tree_complexity(tree) for tree in all_trees)

def knn_complexity(knn):
    return knn.n_neighbors


def support_vector_machine_complexity(svm):
    return sum(svm.n_support_)

def sklearn_MLP_complexity(mlp):
    n_layers = len(mlp.coefs_)
    n_params = 0
    for i in range(n_layers):
        n_params += len(mlp.coefs_[i]) + len(mlp.intercepts_[i])
    return n_params

# TODO use the complexity defined by XGBoost?
def calculate_xgb_model_complexity(model):
    num_nodes = len(model.get_booster().trees_to_dataframe())
    return num_nodes*5

def BernoulliNB_Complexity(model):
    num_coefficients = len(model.class_log_prior_) + len(model.feature_log_prob_)
    return num_coefficients

def GaussianNB_Complexity(model):
    num_coefficients = len(model.class_prior_) + len(model.theta_) + len(model.var_)
    return num_coefficients

def MultinomialNB_Complexity(model):
    num_coefficients = len(model.class_log_prior_) + len(model.feature_log_prob_)
    return num_coefficients

complexity_objective_per_estimator =    {   LogisticRegression: _count_nonzero_coefficients_and_intercept,
                                            RandomForestClassifier : forest_complexity,
                                            SGDClassifier: _count_nonzero_coefficients_and_intercept,
                                            KNeighborsClassifier: knn_complexity,
                                            DecisionTreeClassifier: tree_complexity,
                                            GradientBoostingClassifier: forest_complexity,
                                            ExtraTreesClassifier: forest_complexity,
                                            XGBClassifier: calculate_xgb_model_complexity,
                                            SVC : support_vector_machine_complexity,
                                            LinearSVC : _count_nonzero_coefficients_and_intercept,
                                            MLPClassifier: sklearn_MLP_complexity,
                                            BernoulliNB: BernoulliNB_Complexity,
                                            GaussianNB: GaussianNB_Complexity,
                                        }


def calculate_model_complexity(est):
    if isinstance(est, sklearn.pipeline.Pipeline) or isinstance(est, sklearn.pipeline.FeatureUnion):
        return sum(calculate_model_complexity(estimator) for _,estimator in est.steps)
    if isinstance(est, GraphPipeline):
        return sum(calculate_model_complexity(est.graph.nodes[node]['instance']) for node in est.graph.nodes)

    model_type = type(est)
    if model_type in complexity_objective_per_estimator:
        return complexity_objective_per_estimator[model_type](est)
    #else, if is subclass of sklearn selector
    elif issubclass(model_type, sklearn.feature_selection.SelectorMixin):
        return 0
    else:
        return 1


def complexity_scorer(est, X, y):
    return calculate_model_complexity(est)




from abc import abstractmethod
import types
import numpy as np
import copy
import copy
import typing


class BaseIndividual:


    def __init__(self) -> None:
        self.mutation_list = []
        self.crossover_list = []

    def mutate(self, rng_=None):
        rng = np.random.default_rng(rng_)
        mutation_list_copy = self.mutation_list.copy()
        rng.shuffle(mutation_list_copy)
        for func in mutation_list_copy:
            if func():
                return True
        return False

    def crossover(self, ind2, rng_=None):
        rng = np.random.default_rng(rng_)
        crossover_list_copy = self.crossover_list.copy()
        rng.shuffle(crossover_list_copy)
        for func in crossover_list_copy:
            if func(ind2):
                return True
        return False

    # a guided change of an individual when given an objective function
    def optimize(self, objective_function, rng_=None , steps=5):
        rng = np.random.default_rng(rng_)
        for _ in range(steps):
            self.mutate(rng_=rng)

    #Return a hashable unique to this individual setup
    #For use when evaluating whether or not an individual is 'the same' and another individual
    def unique_id(self):
        return self


    #TODO https://www.pythontutorial.net/python-oop/python-__hash__/
    #python hashing and __eq__ functions look into
    #whether or not this would be a better way of doing things

    # #TODO: use this instead of unique_id()?
    # #unique_id() and __repr__ could have different levels of specificity.
    # def __repr__(self) -> str:
    #     pass

    # def __hash__(self) -> int:
    #     pass

    # def __eq__(self, other):
    #     self.unique_id() == other.unique_id()



import numpy as np

def lexicase_selection(scores, k, rng_=None, n_parents=1,):
    """Select the best individual according to Lexicase Selection, *k* times.
    The returned list contains the indices of the chosen *individuals*.
    :param scores: The score matrix, where rows the individulas and the columns are the corresponds to scores on different objectives.
    :returns: A list of indices of selected individuals.
    This function uses the :func:`~random.choice` function from the python base
    :mod:`random` module.
    """
    rng = np.random.default_rng(rng_)
    chosen =[]
    for i in range(k*n_parents):
        candidates = list(range(len(scores)))
        cases = list(range(len(scores[0])))
        rng.shuffle(cases)

        while len(cases) > 0 and len(candidates) > 1:
            best_val_for_case = max(scores[candidates,cases[0]])
            candidates = [x for x in candidates if scores[x, cases[0]] == best_val_for_case]
            cases.pop(0)
        chosen.append(rng.choice(candidates))

    return np.reshape(chosen, (k, n_parents))


from tpot2.builtin_modules import ArithmeticTransformer, FeatureSetSelector
from functools import partial
import pandas as pd
import numpy as np
from tpot2.builtin_modules import AddTransformer, mul_neg_1_Transformer, MulTransformer, SafeReciprocalTransformer, EQTransformer, NETransformer, GETransformer, GTTransformer, LETransformer, LTTransformer, MinTransformer, MaxTransformer, ZeroTransformer, OneTransformer, NTransformer

# ArithmeticTransformer
def params_arthmetic_operator(trial, name=None):
        return {
                'function': trial.suggest_categorical(f'function_{name}', ["add", "mul_neg_1", "mul", "safe_reciprocal", "eq","ne","ge","gt","le","lt", "min","max","0","1"]),
                }

def make_arithmetic_transformer_config_dictionary():
        return  {
                AddTransformer: {},
                mul_neg_1_Transformer: {},
                MulTransformer: {},
                SafeReciprocalTransformer: {},
                EQTransformer: {},
                NETransformer: {},
                GETransformer: {},
                GTTransformer: {},
                LETransformer: {},
                LTTransformer: {},
                MinTransformer: {},
                MaxTransformer: {},
        }





def params_feature_set_selector(trial, name=None, names_list = None, subset_dict=None):
    """Create a dictionary of parameters for FeatureSetSelector.

    Parameters
    ----------
    trial: optuna.trial.Trial
        A trial corresponds to the evaluation of a objective function.
    name: string
        Used for compatibility in when calling multiple optuna of multiple parameters at once.
    names_list: list of string
        List of names of the feature set selector. To more easily keep track of what the subsets represent.
        Included to prevent repeat calls to list(subset_dict.keys()) which may be slow and/or have different orderings
    subset_dict: dictionary
        A dictionary of subsets. The keys are the names of the subsets and the values are the subsets.

    Returns
    -------
    params: dictionary
        A dictionary of parameters for FeatureSetSelector.
    """

    subset_name = trial.suggest_categorical(f'subset_name_{name}', names_list)

    params =    {'name': subset_name,
                    'sel_subset': subset_dict[subset_name],
                }

    return params

def make_FSS_config_dictionary(subsets=None, n_features=None, feature_names=None):
    """Create the search space of parameters for FeatureSetSelector.

    Parameters
    ----------
    subsets: Sets the subsets to select from.
        - str : If a string, it is assumed to be a path to a csv file with the subsets.
            The first column is assumed to be the name of the subset and the remaining columns are the features in the subset.
        - list or np.ndarray : If a list or np.ndarray, it is assumed to be a list of subsets.

    n_features: int the number of features in the dataset.
        If subsets is None, each column will be treated as a subset. One column will be selected per subset.
    """

    #require at least of of the parameters
    if subsets is None and n_features is None:
        raise ValueError('At least one of the parameters must be provided')

    if isinstance(subsets, str):
        df = pd.read_csv(subsets,header=None,index_col=0)
        df['features'] = df.apply(lambda x: list([x[c] for c in df.columns]),axis=1)
        subset_dict = {}
        for row in df.index:
            subset_dict[row] = df.loc[row]['features']
    elif isinstance(subsets, dict):
        subset_dict = subsets
    elif isinstance(subsets, list) or isinstance(subsets, np.ndarray):
        subset_dict = {str(i):subsets[i] for i in range(len(subsets))}
    else:
        if feature_names is None:
            subset_dict = {str(i):i for i in range(n_features)}
        else:
            subset_dict = {str(i):feature_names[i] for i in range(len(feature_names))}

    names_list = list(subset_dict.keys())

    return {FeatureSetSelector: partial(params_feature_set_selector, names_list = names_list, subset_dict=subset_dict)}



from tpot2.builtin_modules import Passthrough

def params_passthrough(trial, name=None):
    return {}

def make_passthrough_config_dictionary():
    return {Passthrough: params_passthrough}


import sklearn
from sklearn.ensemble import ExtraTreesClassifier, ExtraTreesRegressor
from sklearn.feature_selection import SelectFromModel
from sklearn.feature_selection import RFE

class RFE_ExtraTreesClassifier(RFE):
    def __init__(self, n_features_to_select=None, step=1, verbose=0, importance_getter='auto', n_estimators=100, *, criterion='gini', max_depth=None, min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features='sqrt', max_leaf_nodes=None, min_impurity_decrease=0.0, bootstrap=False, oob_score=False, n_jobs=None, random_state=None, warm_start=False, class_weight=None, ccp_alpha=0.0, max_samples=None):
        
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.warm_start = warm_start
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples


        estimator = ExtraTreesClassifier(n_estimators=n_estimators, criterion=criterion, max_depth=max_depth, min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf, min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features, max_leaf_nodes=max_leaf_nodes, min_impurity_decrease=min_impurity_decrease, bootstrap=bootstrap, oob_score=oob_score, n_jobs=n_jobs, random_state=random_state, verbose=verbose, warm_start=warm_start, class_weight=class_weight, ccp_alpha=ccp_alpha, max_samples=max_samples)
        super().__init__(estimator=estimator, n_features_to_select=n_features_to_select, step=step, verbose=verbose, importance_getter=importance_getter)


class SelectFromModel_ExtraTreesClassifier(SelectFromModel):
    def __init__(self, threshold=None, prefit=False, norm_order=1, SFM_max_features=None, importance_getter='auto', n_estimators=100, *, criterion='gini', max_depth=None, min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features='sqrt', max_leaf_nodes=None, min_impurity_decrease=0.0, bootstrap=False, oob_score=False, n_jobs=None, random_state=None, verbose=0, warm_start=False, class_weight=None, ccp_alpha=0.0, max_samples=None):
        
        self.SFM_max_features = SFM_max_features
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.warm_start = warm_start
        self.class_weight = class_weight
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples
        
        estimator = ExtraTreesClassifier(n_estimators=n_estimators, criterion=criterion, max_depth=max_depth, min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf, min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features, max_leaf_nodes=max_leaf_nodes, min_impurity_decrease=min_impurity_decrease, bootstrap=bootstrap, oob_score=oob_score, n_jobs=n_jobs, random_state=random_state, verbose=verbose, warm_start=warm_start, class_weight=class_weight, ccp_alpha=ccp_alpha, max_samples=max_samples)
        super().__init__(estimator=estimator, threshold=threshold, prefit=prefit, norm_order=norm_order, max_features=SFM_max_features, importance_getter=importance_getter)


class RFE_ExtraTreesRegressor(RFE):
    def __init__(self, n_features_to_select=None, step=1, verbose=0, importance_getter='auto',n_estimators=100, *, criterion='squared_error', max_depth=None, min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features=1.0, max_leaf_nodes=None, min_impurity_decrease=0.0, bootstrap=False, oob_score=False, n_jobs=None, random_state=None, warm_start=False, ccp_alpha=0.0, max_samples=None ) -> None:
        
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.warm_start = warm_start
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples


        
        estimator = ExtraTreesRegressor(n_estimators=n_estimators, criterion=criterion, max_depth=max_depth, min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf, min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features, max_leaf_nodes=max_leaf_nodes, min_impurity_decrease=min_impurity_decrease, bootstrap=bootstrap, oob_score=oob_score, n_jobs=n_jobs, random_state=random_state, verbose=verbose, warm_start=warm_start, ccp_alpha=ccp_alpha, max_samples=max_samples)
        super().__init__(estimator=estimator, n_features_to_select=n_features_to_select, step=step, verbose=verbose, importance_getter=importance_getter)

class SelectFromModel_ExtraTreesRegressor(SelectFromModel):
    def __init__(self, threshold=None, prefit=False, norm_order=1, SFM_max_features=None, importance_getter='auto', n_estimators=100, *, criterion='squared_error', max_depth=None, min_samples_split=2, min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features=1.0, max_leaf_nodes=None, min_impurity_decrease=0.0, bootstrap=False, oob_score=False, n_jobs=None, random_state=None, verbose=0, warm_start=False, ccp_alpha=0.0, max_samples=None ):
        
        self.SFM_max_features = SFM_max_features
        self.n_estimators = n_estimators
        self.criterion = criterion
        self.max_depth = max_depth
        self.min_samples_split = min_samples_split
        self.min_samples_leaf = min_samples_leaf
        self.min_weight_fraction_leaf = min_weight_fraction_leaf
        self.max_features = max_features
        self.max_leaf_nodes = max_leaf_nodes
        self.min_impurity_decrease = min_impurity_decrease
        self.bootstrap = bootstrap
        self.oob_score = oob_score
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.verbose = verbose
        self.warm_start = warm_start
        self.ccp_alpha = ccp_alpha
        self.max_samples = max_samples
        
        estimator = ExtraTreesRegressor(n_estimators=n_estimators, criterion=criterion, max_depth=max_depth, min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf, min_weight_fraction_leaf=min_weight_fraction_leaf, max_features=max_features, max_leaf_nodes=max_leaf_nodes, min_impurity_decrease=min_impurity_decrease, bootstrap=bootstrap, oob_score=oob_score, n_jobs=n_jobs, random_state=random_state, verbose=verbose, warm_start=warm_start, ccp_alpha=ccp_alpha, max_samples=max_samples)
        super().__init__(estimator=estimator, threshold=threshold, prefit=prefit, norm_order=norm_order, max_features=SFM_max_features, importance_getter=importance_getter)



#TODO: how to best support transformers/selectors that take other transformers with their own hyperparameters?
import numpy as np
from sklearn.feature_selection import SelectFwe
from sklearn.feature_selection import SelectPercentile
from sklearn.feature_selection import VarianceThreshold
from sklearn.feature_selection import RFE
from sklearn.feature_selection import SelectFromModel
import sklearn.feature_selection
from functools import partial
from sklearn.ensemble import ExtraTreesRegressor, ExtraTreesClassifier
from tpot2.builtin_modules import RFE_ExtraTreesClassifier, SelectFromModel_ExtraTreesClassifier, RFE_ExtraTreesRegressor, SelectFromModel_ExtraTreesRegressor

from .classifiers import params_ExtraTreesClassifier
from .regressors import params_ExtraTreesRegressor

def params_sklearn_feature_selection_SelectFwe(trial, name=None):
    return {
        'alpha': trial.suggest_float(f'alpha_{name}', 1e-4, 0.05, log=True),
        'score_func' : sklearn.feature_selection.f_classif,
    }

def params_sklearn_feature_selection_SelectPercentile(trial, name=None):
    return {
        'percentile': trial.suggest_float(f'percentile_{name}', 1, 100.0),
        'score_func' : sklearn.feature_selection.f_classif,
    }

def params_sklearn_feature_selection_VarianceThreshold(trial, name=None):
    return {
        'threshold': trial.suggest_float(f'threshold_{name}', 1e-4, .2, log=True)
    }


#TODO add more estimator options? How will that interact with optuna?
def params_sklearn_feature_selection_RFE(trial, random_state=None, name=None, classifier=True):

    if classifier:
        estimator = ExtraTreesClassifier(**params_ExtraTreesClassifier(trial, random_state=random_state, name=f"RFE_{name}"))
    else:
        estimator = ExtraTreesRegressor(**params_ExtraTreesRegressor(trial, random_state=random_state, name=f"RFE_{name}"))

    params = {
            'step': trial.suggest_float(f'step_{name}', 1e-4, 1.0, log=False),
            'estimator' : estimator,
            }

    return params


def params_sklearn_feature_selection_SelectFromModel(trial, random_state=None, name=None, classifier=True):

    if classifier:
        estimator = ExtraTreesClassifier(**params_ExtraTreesClassifier(trial, random_state=random_state, name=f"SFM_{name}"))
    else:
        estimator = ExtraTreesRegressor(**params_ExtraTreesRegressor(trial, random_state=random_state, name=f"SFM_{name}"))

    params = {
            'threshold': trial.suggest_float(f'threshold_{name}', 1e-4, 1.0, log=True),
            'estimator' : estimator,
            }

    return params



def params_sklearn_feature_selection_RFE_wrapped(trial, random_state=None, name=None, classifier=True):

    params = {
            'step': trial.suggest_float(f'step_{name}', 1e-4, 1.0, log=False),
            }

    if classifier:
        estimator_params = params_ExtraTreesClassifier(trial, random_state=random_state, name=f"RFE_{name}")
    else:
        estimator_params = params_ExtraTreesRegressor(trial, random_state=random_state, name=f"RFE_{name}")

    params.update(estimator_params)

    return params


def params_sklearn_feature_selection_SelectFromModel_wrapped(trial, random_state=None, name=None, classifier=True):

    params = {
        'threshold': trial.suggest_float(f'threshold_{name}', 1e-4, 1.0, log=True),
        }

    if classifier:
        estimator_params = params_ExtraTreesClassifier(trial, random_state=random_state, name=f"SFM_{name}")
    else:
        estimator_params = params_ExtraTreesRegressor(trial, random_state=random_state, name=f"SFM_{name}")

    params.update(estimator_params)

    return params



def make_selector_config_dictionary(random_state=None, classifier=True):
    if classifier:
        params =    {RFE_ExtraTreesClassifier : partial(params_sklearn_feature_selection_RFE_wrapped, random_state=random_state, classifier=classifier),
                    SelectFromModel_ExtraTreesClassifier : partial(params_sklearn_feature_selection_SelectFromModel_wrapped, random_state=random_state, classifier=classifier),
                    }
    else:
        params =    {RFE_ExtraTreesRegressor : partial(params_sklearn_feature_selection_RFE_wrapped, random_state=random_state, classifier=classifier),
                    SelectFromModel_ExtraTreesRegressor : partial(params_sklearn_feature_selection_SelectFromModel_wrapped, random_state=random_state, classifier=classifier),
                    }

    params.update({ SelectFwe: params_sklearn_feature_selection_SelectFwe,
                    SelectPercentile: params_sklearn_feature_selection_SelectPercentile,
                    VarianceThreshold: params_sklearn_feature_selection_VarianceThreshold,})

    return params


import pytest
import tpot2
from sklearn.datasets import load_iris
import random
import sklearn

#standard test
@pytest.fixture
def tpot_estimator():
    return tpot2.TPOTEstimator(  population_size=50,
                            scorers=['roc_auc_ovr'],
                            scorers_weights=[1],
                            classification=True,
                            n_jobs=1, 
                            early_stop=5,
                            other_objective_functions= [],
                            other_objective_functions_weights=[],
                            max_time_seconds=300,
                            verbose=3)

@pytest.fixture
def sample_dataset():
    X_train, y_train = load_iris(return_X_y=True)
    return X_train, y_train

def test_tpot_estimator_fit(tpot_estimator,sample_dataset):
    #load iris dataset
    X_train = sample_dataset[0]
    y_train = sample_dataset[1]
    tpot_estimator.fit(X_train, y_train)
    assert tpot_estimator.fitted_pipeline_ is not None

@pytest.fixture
def tpot_estimator_with_pipeline(tpot_estimator,sample_dataset):
    tpot_estimator.fit(sample_dataset[0], sample_dataset[1])
    return tpot_estimator

def test_tpot_estimator_predict(tpot_estimator_with_pipeline,sample_dataset):
    #X_test = [[1, 2, 3], [4, 5, 6]]
    X_test = sample_dataset[0]
    y_pred = tpot_estimator_with_pipeline.predict(X_test)
    assert len(y_pred) == len(X_test)

@pytest.mark.skip(reason="not an informative test. X_test is a list instead of a numpy array or pandas dataframe.")
def test_tpot_estimator_score(tpot_estimator_with_pipeline,sample_dataset):
    random.seed(42)
    #random sample 10% of the dataset
    X_test = random.sample(list(sample_dataset[0]), int(len(sample_dataset[0])*0.1))
    y_test = random.sample(list(sample_dataset[1]), int(len(sample_dataset[1])*0.1))
    scorer = sklearn.metrics.get_scorer('roc_auc_ovo')
    assert isinstance(scorer(tpot_estimator_with_pipeline, X_test, y_test), float)

def test_tpot_estimator_generations_type():
    with pytest.raises(TypeError):
        tpot2.TPOTEstimator(generations="two", population_size=10, verbosity=2)

def test_tpot_estimator_population_size_type():
    with pytest.raises(TypeError):
        tpot2.TPOTEstimator(generations=2, population_size='ten', verbosity=2)

def test_tpot_estimator_verbosity_type():
    with pytest.raises(TypeError):
        tpot2.TPOTEstimator(generations=2, population_size=10, verbosity='high')

def test_tpot_estimator_scoring_type():
    with pytest.raises(TypeError):
        tpot2.TPOTEstimator(generations=2, population_size=10, verbosity=2, scoring=0.5)

def test_tpot_estimator_cv_type():
    with pytest.raises(TypeError):
        tpot2.TPOTEstimator(generations=2, population_size=10, verbosity=2, cv='kfold')

def test_tpot_estimator_n_jobs_type():
    with pytest.raises(TypeError):
        tpot2.TPOTEstimator(generations=2, population_size=10, verbosity=2, n_jobs='all')

def test_tpot_estimator_config_dict_type():
    with pytest.raises(TypeError):
        tpot2.TPOTEstimator(generations=2, population_size=10, verbosity=2, config_dict='config')



@pytest.fixture
def tpot_classifier():
    return tpot2.tpot_estimator.templates.TPOTClassifier(max_time_seconds=300,verbose=3)

@pytest.fixture
def tpot_regressor():
    return tpot2.tpot_estimator.templates.TPOTRegressor(max_time_seconds=300,verbose=3)

def test_tpot_classifier_fit(tpot_classifier,sample_dataset):
    #load iris dataset
    X_train = sample_dataset[0]
    y_train = sample_dataset[1]
    tpot_classifier.fit(X_train, y_train)
    assert tpot_classifier.fitted_pipeline_ is not None

def test_tpot_regressor_fit(tpot_regressor):

    scorer = sklearn.metrics.get_scorer('neg_mean_squared_error')
    X, y = sklearn.datasets.load_diabetes(return_X_y=True)
    X_train, X_test, y_train, y_test = sklearn.model_selection.train_test_split(X, y, train_size=0.75, test_size=0.25)
    tpot_regressor.fit(X_train, y_train)
    assert tpot_regressor.fitted_pipeline_ is not None




from sklearn.linear_model import SGDClassifier
from sklearn.ensemble import RandomForestClassifier, ExtraTreesClassifier, GradientBoostingClassifier
from sklearn.neural_network import MLPClassifier
from sklearn.tree import DecisionTreeClassifier
from xgboost import XGBClassifier
from sklearn.neighbors import KNeighborsClassifier
from sklearn.svm import SVC
from sklearn.linear_model import LogisticRegression
from lightgbm import LGBMClassifier
from sklearn.svm import LinearSVC

from functools import partial
#import GaussianNB

from sklearn.naive_bayes import GaussianNB, BernoulliNB, MultinomialNB

import numpy as np



def params_LogisticRegression(trial, random_state=None, name=None):
    params = {}
    params['solver'] = trial.suggest_categorical(name=f'solver_{name}',
                                                 choices=[f'newton-cg', 'lbfgs', 'liblinear', 'sag', 'saga'])
    params['dual'] = False
    params['penalty'] = 'l2'
    params['C'] = trial.suggest_float(f'C_{name}', 1e-4, 1e4, log=True)
    params['l1_ratio'] = None
    if params['solver'] == 'liblinear':
        params['penalty'] = trial.suggest_categorical(name=f'penalty_{name}', choices=['l1', 'l2'])
        if params['penalty'] == 'l2':
            params['dual'] = trial.suggest_categorical(name=f'dual_{name}', choices=[True, False])
        else:
            params['penalty'] = 'l1'

    params['class_weight'] = trial.suggest_categorical(name=f'class_weight_{name}', choices=['balanced'])
    param_grid = {'solver': params['solver'],
                  'penalty': params['penalty'],
                  'dual': params['dual'],
                  'multi_class': 'auto',
                  'l1_ratio': params['l1_ratio'],
                  'C': params['C'],
                  'n_jobs': 1,
                  'max_iter': 1000,
                  'random_state': random_state
                  }
    return param_grid


def params_KNeighborsClassifier(trial, name=None, n_samples=10):
    return {
        'n_neighbors': trial.suggest_int(f'n_neighbors_{name}', 1, n_samples, log=True ),
        'weights': trial.suggest_categorical(f'weights_{name}', ['uniform', 'distance']),
        'p': trial.suggest_int('p', 1, 3),
        'metric': str(trial.suggest_categorical(f'metric_{name}', ['euclidean', 'minkowski'])),
        'n_jobs': 1,
    }


def params_DecisionTreeClassifier(trial, random_state=None, name=None):
    return {
        'criterion': trial.suggest_categorical(f'criterion_{name}', ['gini', 'entropy']),
        'max_depth': trial.suggest_int(f'max_depth_{name}', 1, 11),
        # 'max_depth_factor' : trial.suggest_float(f'max_depth_factor_{name}', 0, 2, step=0.1),
        'min_samples_split': trial.suggest_int(f'min_samples_split_{name}', 2, 21),
        'min_samples_leaf': trial.suggest_int(f'min_samples_leaf_{name}', 1, 21),
        'min_weight_fraction_leaf': 0.0,
        'max_features': trial.suggest_categorical(f'max_features_{name}', [ 'sqrt', 'log2']),
        'max_leaf_nodes': None,
        'random_state': random_state
    }


def params_SVC(trial, random_state=None, name=None):
    return {
        'kernel': trial.suggest_categorical(name=f'kernel_{name}', choices=['poly', 'rbf', 'linear', 'sigmoid']),
        'C': trial.suggest_float(f'C_{name}', 1e-4, 25, log=True),
        #'gamma': trial.suggest_categorical(name='fgamma_{name}', choices=['scale', 'auto']),
        'degree': trial.suggest_int(f'degree_{name}', 1, 4),
        'class_weight': trial.suggest_categorical(name=f'class_weight_{name}', choices=[None, 'balanced']),
        #'coef0': trial.suggest_float(f'coef0_{name}', 0, 10, step=0.1),
        'max_iter': 3000,
        'tol': 0.005,
        'probability': True,
        'random_state': random_state
    }


def params_LinearSVC(trial, random_state=None, name=None):

    penalty = trial.suggest_categorical(name=f'penalty_{name}', choices=['l1', 'l2'])
    if penalty == 'l1':
        loss = 'squared_hinge'
    else:
        loss = trial.suggest_categorical(name=f'loss_{name}', choices=['hinge', 'squared_hinge'])

    if loss == 'hinge' and penalty == 'l2':
        dual = True
    else:
        dual = trial.suggest_categorical(name=f'dual_{name}', choices=[True, False])

    return {
        'penalty': penalty,
        'loss': loss,
        'dual': dual,
        'C': trial.suggest_float(f'C_{name}', 1e-4, 25, log=True),
        'random_state': random_state
    }


def params_RandomForestClassifier(trial, random_state=None, name=None):
    params = {
        'n_estimators': 100,
        'criterion': trial.suggest_categorical(name=f'criterion_{name}', choices=['gini', 'entropy']),
        #'max_features': trial.suggest_categorical('max_features_{name}', ['auto', 'sqrt', 'log2']),
        'bootstrap': trial.suggest_categorical(name=f'bootstrap_{name}', choices=[True, False]),
        'min_samples_split': trial.suggest_int(f'min_samples_split_{name}', 2, 20),
        'min_samples_leaf': trial.suggest_int(f'min_samples_leaf_{name}', 1, 20),
        'n_jobs': 1,
        'random_state': random_state
    }
    return params


def params_GradientBoostingClassifier(trial, random_state=None, n_classes=None, name=None):

    if n_classes is not None and n_classes > 2:
        loss = 'log_loss'
    else:
        loss = trial.suggest_categorical(name=f'loss_{name}', choices=['log_loss', 'exponential'])

    params = {
        'n_estimators': 100,
        'loss': loss,
        'learning_rate': trial.suggest_float(f'learning_rate_{name}', 1e-3, 1, log=True),
        'min_samples_leaf': trial.suggest_int(f'min_samples_leaf_{name}', 1, 20),
        'min_samples_split': trial.suggest_int(f'min_samples_split_{name}', 2, 20),
        'subsample': trial.suggest_float(f'subsample_{name}', 0.1, 1.0),
        'max_features': trial.suggest_float(f'max_features_{name}', 0.1, 1.0),
        'max_depth': trial.suggest_int(f'max_depth_{name}', 1, 10),
        'tol': 1e-4,
        'random_state': random_state
    }
    return params


def params_XGBClassifier(trial, random_state=None, name=None):
    return {
        'learning_rate': trial.suggest_float(f'learning_rate_{name}', 1e-3, 1, log=True),
        'subsample': trial.suggest_float(f'subsample_{name}', 0.1, 1.0),
        'min_child_weight': trial.suggest_int(f'min_child_weight_{name}', 1, 21),
        #'booster': trial.suggest_categorical(name='booster_{name}', choices=['gbtree', 'dart']),
        'n_estimators': 100,
        'max_depth': trial.suggest_int(f'max_depth_{name}', 1, 11),
        'n_jobs': 1,
        #'use_label_encoder' : True,
        'random_state': random_state
    }


def params_LGBMClassifier(trial, random_state=None, name=None):
    params = {
        'objective': 'binary',
        'metric': 'binary_logloss',
        'boosting_type': trial.suggest_categorical(name=f'boosting_type_{name}', choices=['gbdt', 'dart', 'goss']),
        'num_leaves': trial.suggest_int(f'num_leaves_{name}', 2, 256),
        'max_depth': trial.suggest_int(f'max_depth_{name}', 1, 10),
        'n_estimators': trial.suggest_int(f'n_estimators_{name}', 10, 100),  # 200-6000 by 200
        'deterministic': True,
        'force_row_wise': True,
        'n_jobs': 1,
        'random_state': random_state

    }
    if 2 ** params['max_depth'] > params['num_leaves']:
        params['num_leaves'] = 2 ** params['max_depth']
    return params


def params_ExtraTreesClassifier(trial, random_state=None, name=None):
    params = {
        'n_estimators': 100,
        'criterion': trial.suggest_categorical(name=f'criterion_{name}', choices=["gini", "entropy"]),
        'max_features': trial.suggest_float('max_features', 0.05, 1.00),
        'min_samples_split': trial.suggest_int(f'min_samples_split_{name}', 2, 21,step=1),
        'min_samples_leaf': trial.suggest_int(f'min_samples_leaf_{name}', 1, 21, step=1),
        'bootstrap': trial.suggest_categorical(f'bootstrap_{name}', [True, False]),
        'n_jobs': 1,
        'random_state': random_state
    }
    return params

def params_SGDClassifier(trial, random_state=None, name=None):
    params = {
        'loss': trial.suggest_categorical(f'loss_{name}', ['log_loss', 'modified_huber',]),
        'penalty': 'elasticnet',
        'alpha': trial.suggest_float(f'alpha_{name}', 1e-5, 0.01, log=True),
        'learning_rate': trial.suggest_categorical(f'learning_rate_{name}', ['invscaling', 'constant']),
        'fit_intercept': True,
        'l1_ratio': trial.suggest_float(f'l1_ratio_{name}', 0.0, 1.0),
        'eta0': trial.suggest_float(f'eta0_{name}', 0.01, 1.0),
        'power_t': trial.suggest_float(f'power_t_{name}', 1e-5, 100.0, log=True),
        'n_jobs': 1,
        'random_state': random_state
    }

    return params

def params_MLPClassifier_tpot(trial, random_state=None, name=None):
    params = {
        'alpha': trial.suggest_float(f'alpha_{name}', 1e-4, 1e-1, log=True),
        'learning_rate_init': trial.suggest_float(f'learning_rate_init_{name}', 1e-3, 1., log=True),
        'random_state': random_state
    }

    return params

def params_MLPClassifier_large(trial, name=None):
    n_layers = trial.suggest_int(f'n_layers_{name}', 2, 3)
    layers = []
    for i in range(n_layers):
        layers.append(trial.suggest_int(f'n_neurons_{i}_{name}', 4, 128))

    params = {
        'activation':  trial.suggest_categorical(name=f'activation_{name}', choices=['identity', 'logistic', 'tanh', 'relu']),
        'solver':  trial.suggest_categorical(name=f'solver_{name}', choices=['lbfgs', 'sgd', 'adam']),
        'alpha':  trial.suggest_float(f'alpha_{name}', 0.0001, 1.0, log=True),
        'hidden_layer_sizes':  tuple(layers),
        'max_iter' : 10000
    }

    return params


def params_BernoulliNB(trial, name=None):
    params = {
        'alpha': trial.suggest_float(f'alpha_{name}', 1e-3, 100, log=True),
        'fit_prior': trial.suggest_categorical(f'fit_prior_{name}', [True, False]),
    }
    return params


def params_MultinomialNB(trial, name=None):
    params = {
        'alpha': trial.suggest_float(f'alpha_{name}', 1e-3, 100, log=True),
        'fit_prior': trial.suggest_categorical(f'fit_prior_{name}', [True, False]),
    }
    return params


def make_classifier_config_dictionary(random_state=None, n_samples=10, n_classes=None):
    n_samples = min(n_samples,100) #TODO optimize this

    return {
            LogisticRegression: partial(params_LogisticRegression, random_state=random_state),
            DecisionTreeClassifier: partial(params_DecisionTreeClassifier, random_state=random_state),
            KNeighborsClassifier:  partial(params_KNeighborsClassifier,n_samples=n_samples),
            GradientBoostingClassifier: partial(params_GradientBoostingClassifier, random_state=random_state, n_classes=n_classes),
            ExtraTreesClassifier: partial(params_ExtraTreesClassifier, random_state=random_state),
            RandomForestClassifier: partial(params_RandomForestClassifier, random_state=random_state),
            SGDClassifier: partial(params_SGDClassifier, random_state=random_state),
            GaussianNB: {},
            BernoulliNB: params_BernoulliNB,
            MultinomialNB: params_MultinomialNB,
            XGBClassifier: partial(params_XGBClassifier, random_state=random_state),
            #LinearSVC: partial(params_LinearSVC, random_state=random_state),
            SVC: partial(params_SVC, random_state=random_state),
            #: params_LGBMClassifier, # logistic regression and SVM/SVC are just special cases of this one? remove?
            MLPClassifier: partial(params_MLPClassifier_tpot, random_state=random_state),
        }



import time
import sklearn.metrics
from collections.abc import Iterable
import pandas as pd
import sklearn
import numpy as np

def cross_val_score_objective(pipeline, X, y, scorers, cv, fold=None):
    #check if scores is not iterable
    if not isinstance(scorers, Iterable): 
        scorers = [scorers]
    scores = []
    if fold is None:
        for train_index, test_index in cv.split(X, y):
            this_fold_pipeline = sklearn.base.clone(pipeline)
            if isinstance(X, pd.DataFrame) or isinstance(X, pd.Series):
                X_train, X_test = X.iloc[train_index], X.iloc[test_index]
            else:
                X_train, X_test = X[train_index], X[test_index]

            if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
                y_train, y_test = y.iloc[train_index], y.iloc[test_index]
            else:
                y_train, y_test = y[train_index], y[test_index]


            start = time.time()
            this_fold_pipeline.fit(X_train,y_train)
            duration = time.time() - start

            this_fold_scores = [sklearn.metrics.get_scorer(scorer)(this_fold_pipeline, X_test, y_test) for scorer in scorers] 
            scores.append(this_fold_scores)
            del this_fold_pipeline
            del X_train
            del X_test
            del y_train
            del y_test
            

        return np.mean(scores,0)
    else:
        this_fold_pipeline = sklearn.base.clone(pipeline)
        train_index, test_index = list(cv.split(X, y))[fold]
        if isinstance(X, pd.DataFrame) or isinstance(X, pd.Series):
            X_train, X_test = X.iloc[train_index], X.iloc[test_index]
        else:
            X_train, X_test = X[train_index], X[test_index]

        if isinstance(y, pd.DataFrame) or isinstance(y, pd.Series):
            y_train, y_test = y.iloc[train_index], y.iloc[test_index]
        else:
            y_train, y_test = y[train_index], y[test_index]

        start = time.time()
        this_fold_pipeline.fit(X_train,y_train)
        duration = time.time() - start
        this_fold_scores = [sklearn.metrics.get_scorer(scorer)(this_fold_pipeline, X_test, y_test) for scorer in scorers] 
        return this_fold_scores









